# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Function for loading AndroidEnv."""

import os
import os.path

from android_env import environment
from android_env.components import coordinator as coordinator_lib
from android_env.components import task_manager as task_manager_lib
from android_env.components.simulators.emulator import emulator_simulator
from android_env.proto import task_pb2
from android_env.components.tools.types import TextModel, IconModel
from android_env.components.tools import naive_functions
from android_env.utils import fix_path

from google.protobuf import text_format

from typing import Optional
from typing import Dict
import functools

def load( task_path: str
        , avd_name: str
        , android_avd_home: str = '~/.android/avd'
        , android_sdk_root: str = '~/Android/Sdk'
        , emulator_path: str = '~/Android/Sdk/emulator/emulator'
        , adb_path: str = '~/Android/Sdk/platform-tools/adb'
        , run_headless: bool = False
        , mitm_config: Optional[Dict[str, str]] = None
        , start_token_mark: str = ""
        , non_start_token_mark: str = "##"
        , special_token_pattern: str = r"\[\w+\]"
        , unify_vocabulary: Optional[str] = None
        , text_model: TextModel = naive_functions
        , icon_model: IconModel = naive_functions
        , with_view_hierarchy: bool = False
        ) -> environment.AndroidEnv:
  """Loads an AndroidEnv instance.

  Args:
    task_path: Path to the task textproto file or the directory of task
      textproto file set
    avd_name: Name of the AVD (Android Virtual Device).

    android_avd_home: Path to the AVD (Android Virtual Device).
    android_sdk_root: Root directory of the SDK.
    emulator_path: Path to the emulator binary.
    adb_path: Path to the ADB (Android Debug Bridge).
    run_headless: If True, the emulator display is turned off.
    mitm_config: An optional dict to config the launch options if an mitm proxy
      is in need for a web-based app. The dict is expected to have a struture
      like
        {
            "method": "syscert" | "frida" | "packpatch",
            "address": str as the address for the mitm proxy server, default to
              "127.0.0.1",
            "port": str as the port for the mitm proxy server, default to
              "8080",
            "frida-server": an option for method "frida", to indicate the path
              to the executable of frida-server on the android system, default
              to "/data/local/tmp/frida-server",
            "frida": an option for method "frida", to indicate the path to the
              executable of frida on the local machine, default to "frida",
            "frida-script": an option for method "frida", to indicate the
              script used to hijack the target applicationa, default to
              "frida-script.js",
            "patch-suffix": an option for method "packpatch", to indicate the
              suffix $suff with which the package name "$package.apk" to be
              installed is replaced to "$package-$suff.apk", default to
              "patched"
        }

    start_token_mark: str as the mark for the start subwords such as "Ġ"
      (\\u0120) for GPT subword
    non_start_token_mark: str as the mark for the non-start subwords such as
      "##" for BERT subword
    special_token_pattern: str as the pattern of the special non-printable
      tokens such as "[CLS]" for BERT tokens
    unify_vocabulary: str or none as a file name to the vocabulary file in
      which each line constitutes a token.

    text_model (TextModel): text model to conduct text recognition on the
      screen
    icon_model (IconModel): icon model to conduct icon recognition on the
      screen
    with_view_hierarchy (bool): if the view hierarchy should be included in the
      observation

  Returns:
    env: An AndroidEnv instance.
  """

  adb_controller_args = dict(
      adb_path=os.path.expanduser(adb_path),
      adb_server_port=5037,
      prompt_regex=r'\w*:\/ [$#]')
  emulator_launcher_args = dict(
      avd_name=avd_name,
      android_avd_home=os.path.expanduser(android_avd_home),
      android_sdk_root=os.path.expanduser(android_sdk_root),
      emulator_path=os.path.expanduser(emulator_path),
      run_headless=run_headless,
      gpu_mode='swiftshader_indirect',
      writable_system=False)

  # Prepare task.
  task_list = []
  if os.path.isdir(task_path):
    task_paths = list( map( functools.partial(os.path.join, task_path)
                          , sorted(
                              filter( lambda f: f.endswith(".textproto")
                                    , os.listdir(task_path)
                                    )
                              )
                          )
                     )
    task_directory = task_path
  else:
    task_paths = [task_path]
    task_directory = os.path.dirname(task_path)
  for t_p in task_paths:
    task = task_pb2.Task()
    with open(t_p, 'r') as proto_file:
      text_format.Parse(proto_file.read(), task)
    task_list.append(task)

  adb_root = False
  frida_server = None
  if mitm_config is not None:
    emulator_launcher_args["writable_system"] = True
    emulator_launcher_args["proxy_address"] = mitm_config.get("address", "127.0.0.1")
    emulator_launcher_args["proxy_port"] = mitm_config.get("port", "8080")

    if mitm_config["method"]=="frida":
      adb_root = True
      frida_server = mitm_config.get("frida-server",
          "/data/local/tmp/frida-server")
      adb_controller_args["frida"] = mitm_config.get("frida", "frida")
      adb_controller_args["frida_script"] = mitm_config.get("frida-script",
          "frida-script.js")
    elif mitm_config["method"]=="packpatch":
      for t in task_list:
        for st in t.setup_steps:
          if st.HasField("adb_call") and\
              st.adb_call.HasField("install_apk"):
            apk_path = st.adb_call.install_apk.filesystem.path
            main_name, extension = os.path.splitext(apk_path)
            st.adb_call.install_apk.filesystem.path =\
                "{:}-{:}{:}".format(main_name,
                  mitm_config.get("patch-suffix", "patched"),
                  extension)
    #elif mitm_config["method"]=="syscert":
      #adb_root = True

  # transform the paths in task definitions
  for t in task_list:
    #for st in t.setup_steps:
      #if st.HasField("adb_call") and\
          #st.adb_call.HasField("install_apk"):
        #apk_path = st.adb_call.install_apk.filesystem.path
        #if not os.path.isabs(apk_path):
          #st.adb_call.install_apk.filesystem.path =\
              #os.path.normpath(os.path.join(task_directory, apk_path))
    fix_path(t, task_directory)

  # Create simulator.
  #print("ZDY: BEFORE Simulator Initialization")
  simulator = emulator_simulator.EmulatorSimulator(
      adb_controller_args=adb_controller_args,
      emulator_launcher_args=emulator_launcher_args,
      adb_root=adb_root,
      frida_server=frida_server)
  #print("ZDY: AFTER Simulator Initialization")

  if unify_vocabulary is not None:
    with open(unify_vocabulary) as f:
      vocabulary = list(map(str.strip, f))
  else:
    vocabulary = None

  task_managers = list( map( functools.partial( task_manager_lib.TaskManager
                                              , start_token_mark=start_token_mark
                                              , non_start_token_mark=non_start_token_mark
                                              , special_token_pattern=special_token_pattern
                                              , fix_vocabulary_to=vocabulary
                                              , text_model=text_model
                                              , icon_model=icon_model
                                              )
                           , task_list
                           )
                      )
  coordinator = coordinator_lib.Coordinator( simulator, task_managers
                                           , with_view_hierarchy=with_view_hierarchy
                                           )

  # Load environment.
  return environment.AndroidEnv(coordinator=coordinator)



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

import easyocr
from android_env.components.tools.types import TextModel

from typing import Union, Optional
from typing import List
import threading

import torch
import numpy as np

def in_bbox(bbox0: torch.Tensor, bbox1: torch.Tensor) -> torch.Tensor:
    """
    Test if `bbox0` is contained in `bbox1`.

    Args:
        bbox0 - torch.Tensor of float32 with shape (m, 4)
        bbox1 - torch.Tensor of float32 with shape (n, 4)

    Returns:
        torch.Tensor of bool with shape (n, m)
    """

    bbox0 = bbox0[None, :, :] # (1, m, 4)
    bbox1 = bbox1[:, None, :] # (n, 1, 4)

    left_side = bbox0[..., 0]>=bbox1[..., 0] # (n, m)
    up_side = bbox0[..., 1]>=bbox1[..., 1] # (n, m)
    right_side = bbox0[..., 2]<=bbox1[..., 2] # (n, m)
    down_side = bbox0[..., 3]<=bbox1[..., 3] # (n, m)

    return torch.logical_and( left_side.logical_and(up_side)
                            , right_side.logical_and(down_side)
                            )

class EasyOCRWrapper(TextModel):
    def __init__( self
                , gpu: Union[bool, str, torch.device] = True
                , model_storage_directory: Optional[str] = None
                , download_enabled: bool = True
                ):
        #  method __init__ {{{ # 
        """
        Args:
            gpu: bool or str or torch.device to specify the model device
            model_storage_directory: optional str for the EasyOCR model path.
              EasyOCR will check the following variables in order to determine
              in which directory the model should be found:
              + `model_storage_directory`
              + $EASYOCR_MODULE_PATH
              + $MODULE_PATH
              + ~/.EasyOCR
            download_enabled: bool
        """

        self._reader: easyocr.Reader =\
                easyocr.Reader( ["en"], gpu=gpu
                              , model_storage_directory=model_storage_directory
                              , download_enabled=download_enabled
                              )
        self._lock: threading.Lock = threading.Lock()
        #  }}} method __init__ # 

    def _convert_screen(screen: torch.Tensor) -> np.ndarray:
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width) as
              the RGB image

        Returns:
            np.ndarray of uint8 with shape (height, width, 3) as the BGR image
        """

        screen = screen.cpu().numpy()
        screen = np.clip(screen*255., 0, 255).astype(np.uint8) # convert to uint8
        screen = np.transpose(screen, (1, 2, 0)) # convert (3, H, W) to (H, W, 3)
        screen = screen[..., [2, 1, 0]] # convert RGB to BGR
        return screen

    def text_detector( self
                     , screen: torch.Tensor
                     , bboxes: List[torch.Tensor]
                     ) -> List[List[str]]:
        #  method text_detector {{{ # 
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            bboxes: list with length nb_bboxes of torch.Tensor of float32 with shape (1, 4)

        Returns:
            list with length nb_bboxes of list of str
        """

        screen = self._convert_screen(screen) # (H, W, 3)
        with self._lock:
            results = self._reader.readtext(screen) # get all results

        # check the bboxes
        target_bboxes = torch.cat(bboxes) # (N, 4); N is nb_bboxes
        result_bboxes = torch.tensor(
                            list( map( lambda bb: [bb[0][0], bb[0][1], bb[1][0], bb[2][1]]
                                     , map( lambda rst: rst[0]
                                          , results
                                          )
                                     )
                                )
                            ) # (M, 4); M is the number of results
        region_mask = in_bbox(result_bboxes, target_bboxes) # (N, M)

        result_texts = list( map( lambda rst: rst[1]
                                , results
                                )
                           )
        final_returns = []
        for m in region_mask:
            final_returns.append([t for t, m_ in zip(result_texts, m) if m_])
        return final_returns
        #  }}} method text_detector # 

    def text_recognizer( self
                       , screen: torch.Tensor
                       , bboxes: List[torch.Tensor]
                       ) -> List[str]:
        #  method text_recognizer {{{ # 
        """
        Args:
            screen: torch.Tensor of float32 with shape (3, height, width)
            bboxes: list with length nb_bboxes of torch.Tensor of float32 with shape (1, 4)

        Returns:
            list with length nb_bboxes of str
        """

        screen = self._convert_screen(screen) # (H, W, 3)
        with self._lock:
            results = self._reader.readtext(screen)

        # check the bboxes
        target_bboxes = torch.cat(bboxes) # (N, 4); N is nb_bboxes
        result_bboxes = torch.tensor(
                            list( map( lambda bb: [bb[0][0], bb[0][1], bb[1][0], bb[2][1]]
                                     , map( lambda rst: rst[0]
                                          , results
                                          )
                                     )
                                )
                            ) # (M, 4); M is the number of results
        region_mask = in_bbox(result_bboxes, target_bboxes) # (N, M)

        result_texts = list( map( lambda rst: rst[1]
                                , results
                                )
                           )
        final_returns = []
        for m in region_mask:
            final_returns.append(" ".join((t for t, m_ in zip(result_texts, m) if m_)))
        return final_returns
        #  }}} method text_recognizer # 



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps the AndroidEnv environment to rescale the observations."""

from typing import Optional, Sequence, Dict

from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import numpy as np
from PIL import Image


# Taken from https://pillow.readthedocs.io/en/3.2.x/reference/Image.html#PIL.Image.Image.convert
#
# This array maps an RGB image to a grayscale image using the ITU-R 709
# specification which is good for computer displays and HDTV.
RGB_TO_GRAYSCALE_COEFFICIENTS = [0.2126, 0.7152, 0.0722]


class ImageRescaleWrapper(base_wrapper.BaseWrapper):
  """AndroidEnv with rescaled observations."""

  def __init__(
      self,
      env: dm_env.Environment,
      zoom_factors: Optional[Sequence[float]] = (0.5, 0.5),
      grayscale: bool = False):
    super().__init__(env)
    assert 'pixels' in self._env.observation_spec()
    assert self._env.observation_spec()['pixels'].shape[-1] in [1, 3], (
        'Number of pixel channels should be 1 or 3.')
    self._grayscale = grayscale
    if zoom_factors is None:
      zoom_factors = (1.0, 1.0)
    # We only zoom the width and height of each layer, and we explicitly do not
    # want to zoom the number of channels so we just multiply it by 1.0.
    self._zoom_factors = tuple(zoom_factors) + (1.0,)

    # Save the raw image for making videos, for example.
    self._raw_pixels = None

  @property
  def raw_pixels(self):
    return self._raw_pixels

  def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
    observation = timestep.observation
    self._raw_pixels = observation['pixels'].copy()
    processed_observation = observation.copy()
    processed_observation['pixels'] = self._process_pixels(
        observation['pixels'])
    return timestep._replace(observation=processed_observation)

  def _process_pixels(self, raw_observation: np.ndarray) -> np.ndarray:
    # We expect `raw_observation` to have shape (H, W, 3) - 3 for RGB
    new_shape = np.array(
        self._zoom_factors[0:2] * np.array(raw_observation.shape[0:2]),
        dtype=np.int)[::-1] # ZDY_MARK: PIL.Image.Image.resize accepts (W, H) as the argument
    if self._grayscale:
      # When self._grayscale == True, we squash the RGB into a single layer
      image = np.dot(raw_observation, RGB_TO_GRAYSCALE_COEFFICIENTS)
    else:
      image = raw_observation
    return self._resize_image_array(image, new_shape)

  def _resize_image_array(
      self,
      grayscale_or_rbg_array: np.ndarray,
      new_shape: Sequence[int]) -> np.ndarray:
    """Resize color or grayscale/action_layer array to new_shape."""
    assert np.array(new_shape).ndim == 1
    assert len(new_shape) == 2
    resized_array = np.array(Image.fromarray(
        grayscale_or_rbg_array.astype('uint8')).resize(new_shape))
    if resized_array.ndim == 2:
      return np.expand_dims(resized_array, axis=-1)
    return resized_array

#def reset(self) -> dm_env.TimeStep:
    #timestep = self._env.reset()
    #return self._process_timestep(timestep)

  def step(self, action) -> dm_env.TimeStep:
    timestep = self._env.step(action)
    return self._process_timestep(timestep)

  def observation_spec(self) -> Dict[str, specs.Array]:
    parent_spec = self._env.observation_spec().copy()
    out_shape = np.multiply(parent_spec['pixels'].shape,
                            self._zoom_factors).astype(np.int32)
    if self._grayscale:
      # In grayscale mode we want the output shape to be [H, W, 1]
      out_shape[-1] = 1
    parent_spec['pixels'] = specs.Array(
        shape=out_shape,
        dtype=parent_spec['pixels'].dtype,
        name=parent_spec['pixels'].name)
    return parent_spec



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for fake_simulator."""

from absl.testing import absltest
from android_env.components.simulators.fake import fake_simulator
import numpy as np


class FakeSimulatorTest(absltest.TestCase):

  def test_device_name(self):
    simulator = fake_simulator.FakeSimulator(screen_dimensions=(320, 480))
    self.assertEqual(simulator.adb_device_name(), 'fake_simulator')

  def test_launch_close(self):
    # The simulator should launch and not crash.
    simulator = fake_simulator.FakeSimulator(screen_dimensions=(320, 480))
    simulator.launch()
    # After a successful launch(), screen_dimensions() should return something.
    np.testing.assert_equal(simulator.screen_dimensions(), [320, 480])
    # Closing the simulator should also not crash.
    simulator.close()

  def test_get_observation(self):
    simulator = fake_simulator.FakeSimulator(screen_dimensions=(320, 480))
    simulator.launch()

    observation = simulator.get_observation()
    np.testing.assert_equal(
        observation['pixels'].shape, [320, 480, 3])
    np.testing.assert_array_equal(
        observation['pixels'], np.zeros((320, 480, 3), dtype=np.uint8))
    np.testing.assert_equal(
        observation['timedelta'].dtype, np.int64)
    np.testing.assert_array_equal(
        observation['orientation'], np.zeros((4,), dtype=np.uint8))

  def test_log_stream(self):
    simulator = fake_simulator.FakeSimulator(screen_dimensions=(320, 480))
    simulator.launch()
    log_stream = simulator.get_log_stream()
    lines = [
        '',
        '         1553110400.424  5583  5658 D Tag: reward: 0.5',
        '         1553110400.424  5583  5658 D Tag: reward: 1.0',
        '         1553110400.424  5583  5658 D Tag: extra: my_extra: [1.0]',
        '         1553110400.424  5583  5658 D Tag: episode end',
    ]
    for i, line in enumerate(log_stream.get_stream_output()):
      self.assertIn(line, lines)
      if i > 10:
        break

if __name__ == '__main__':
  absltest.main()



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

from android_env.proto import task_pb2
import os.path
from typing import Optional

def fix_path(task: task_pb2.Task, task_directory: str) -> task_pb2.Task:
    """
    Fixes the path setting in `task` and return it back.

    Args:
        task (task_pb2.Task): task config object
        task_directory (str): the path which is expected to be the base path,
          usually is the directory of the task config file

    Return:
        task_pb2.Task: task config object with the fixed path
    """

    for st in task.setup_steps:
        _fix(st, task_directory)
    for st in task.reset_steps:
        _fix(st, task_directory)
    for evt_s in task.event_sources:
        if evt_s.HasField("icon_match"):
            attribute_name: Optional[str] = evt_s.WhichOneof("event")
            if attribute_name=="icon_match" or attribute_name=="icon_detect_match":
                path: str = getattr(evt_s, attribute_name).path
                if not os.path.isabs(path):
                    getattr(evt_s, attribute_name).path =\
                            os.path.normpath(os.path.join(task_directory, path))
    return task

def _fix(setup: task_pb2.SetupStep, task_directory: str):
    if setup.HasField("adb_call")\
            and setup.adb_call.HasField("install_apk"):
        apk_path = setup.adb_call.install_apk.filesystem.path
        if not os.path.isabs(apk_path):
            setup.adb_call.install_apk.filesystem.path =\
                    os.path.normpath(os.path.join(task_directory, apk_path))



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.adb_controller."""

import os
import time

from absl.testing import absltest
from absl.testing import parameterized
from android_env.components import adb_controller
from android_env.components import errors
from android_env.proto import adb_pb2
import mock

# Timeout to be used by default in tests below. Set to a small value to avoid
# hanging on a failed test.
_TIMEOUT = 2


class AdbControllerTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._mock_execute_command = self.enter_context(
        mock.patch.object(
            adb_controller.AdbController, '_execute_command', autospec=True))
    self._adb_controller = adb_controller.AdbController(
        adb_path='my_adb',
        device_name='awesome_device',
        adb_server_port=9999,
        prompt_regex='l33t>')

  @mock.patch.object(
      adb_controller.AdbController, '_wait_for_device', autospec=True)
  def test_set_touch_indicators(self, mock_wait_for_device):
    self._adb_controller.set_touch_indicators(
        show_touches=True, pointer_location=False, timeout=_TIMEOUT)
    mock_wait_for_device.assert_called_once()
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller,
                  ['shell', 'settings', 'put', 'system', 'show_touches', '1'],
                  _TIMEOUT),
        mock.call(
            self._adb_controller,
            ['shell', 'settings', 'put', 'system', 'pointer_location', '0'],
            _TIMEOUT)
    ])

  def test_force_stop(self):
    self._adb_controller.force_stop('com.amazing.package', timeout=_TIMEOUT)
    self._mock_execute_command.assert_called_once_with(
        self._adb_controller,
        ['shell', 'am', 'force-stop', 'com.amazing.package'], _TIMEOUT)

  def test_clear_cache(self):
    self._adb_controller.clear_cache('com.amazing.package', timeout=_TIMEOUT)
    self._mock_execute_command.assert_called_with(
        self._adb_controller, ['shell', 'pm', 'clear', 'com.amazing.package'],
        _TIMEOUT)

  def test_grant_permissions(self):
    self._adb_controller.grant_permissions(
        'com.amazing.package', ['hey.READ', 'ho.WRITE'], timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller,
                  ['shell', 'pm', 'grant', 'com.amazing.package', 'hey.READ'],
                  _TIMEOUT),
        mock.call(self._adb_controller,
                  ['shell', 'pm', 'grant', 'com.amazing.package', 'ho.WRITE'],
                  _TIMEOUT),
    ])

  def test_start_activity(self):
    self._adb_controller.start_activity(
        'hello.world/hello.world.MainActivity', [], timeout=_TIMEOUT)
    self._adb_controller.start_activity(
        full_activity='hello.world/hello.world.MainActivity',
        extra_args=['Planet 1', 'Planet 2'],
        timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller, [
            'shell', 'am', 'start', '-S', '-n',
            'hello.world/hello.world.MainActivity'
        ], _TIMEOUT),
        mock.call(self._adb_controller, [
            'shell', 'am', 'start', '-S', '-n',
            'hello.world/hello.world.MainActivity', 'Planet 1', 'Planet 2'
        ], _TIMEOUT)
    ])

  def test_start_intent(self):
    self._adb_controller.start_intent(
        action='action',
        data_uri='data',
        package_name='my.package',
        timeout=_TIMEOUT)

    self._mock_execute_command.assert_called_once_with(
        self._adb_controller,
        ['shell', 'am', 'start', '-a', 'action', '-d', 'data', 'my.package'],
        _TIMEOUT)

  def test_broadcast(self):
    self._adb_controller.broadcast(
        'hello.world/hello.world.BroadcastReceiver',
        'android.intent.action.TEST', [],
        timeout=_TIMEOUT)
    self._adb_controller.broadcast(
        receiver='hello.world/hello.world.BroadcastReceiver',
        action='android.intent.action.TEST',
        extra_args=['--es', 'KEY', 'VALUE'],
        timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller, [
            'shell', 'am', 'broadcast', '-n',
            'hello.world/hello.world.BroadcastReceiver', '-a',
            'android.intent.action.TEST'
        ], _TIMEOUT),
        mock.call(self._adb_controller, [
            'shell', 'am', 'broadcast', '-n',
            'hello.world/hello.world.BroadcastReceiver', '-a',
            'android.intent.action.TEST', '--es', 'KEY', 'VALUE'
        ], _TIMEOUT)
    ])

  def test_setprop(self):
    self._adb_controller.setprop('myprop', 'true', timeout=_TIMEOUT)
    self._adb_controller.setprop('myotherprop', 'false', timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller, ['shell', 'setprop', 'myprop', 'true'],
                  _TIMEOUT),
        mock.call(self._adb_controller,
                  ['shell', 'setprop', 'myotherprop', 'false'], _TIMEOUT),
    ])

  def test_rotate_device(self):
    self._adb_controller.rotate_device(
        adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_90, timeout=_TIMEOUT)
    self._adb_controller.rotate_device(
        adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_0, timeout=_TIMEOUT)
    self._adb_controller.rotate_device(
        adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_270, timeout=_TIMEOUT)
    self._adb_controller.rotate_device(
        adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_180, timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(
            self._adb_controller,
            args=['shell', 'settings', 'put', 'system', 'user_rotation', '1'],
            timeout=_TIMEOUT),
        mock.call(
            self._adb_controller,
            args=['shell', 'settings', 'put', 'system', 'user_rotation', '0'],
            timeout=_TIMEOUT),
        mock.call(
            self._adb_controller,
            args=['shell', 'settings', 'put', 'system', 'user_rotation', '3'],
            timeout=_TIMEOUT),
        mock.call(
            self._adb_controller,
            args=['shell', 'settings', 'put', 'system', 'user_rotation', '2'],
            timeout=_TIMEOUT)
    ])

  @mock.patch.object(
      adb_controller.AdbController, '_wait_for_device', autospec=True)
  def test_get_screen_dimensions_failed_wait(self, mock_wait_for_device):

    mock_wait_for_device.side_effect = errors.AdbControllerDeviceTimeoutError(
        'Time is up.')
    self.assertRaises(errors.AdbControllerDeviceTimeoutError,
                      self._adb_controller.get_screen_dimensions)
    mock_wait_for_device.assert_called_once()

  @mock.patch.object(
      adb_controller.AdbController, '_wait_for_device', autospec=True)
  def test_get_screen_dimensions_success(self, mock_wait_for_device):
    self._mock_execute_command.return_value = b'Physical size: 1280x800'
    screen_dimensions = self._adb_controller.get_screen_dimensions(
        timeout=_TIMEOUT)
    mock_wait_for_device.assert_called()
    self._mock_execute_command.assert_called_with(self._adb_controller,
                                                  ['shell', 'wm', 'size'],
                                                  _TIMEOUT)
    self.assertEqual(screen_dimensions, (800, 1280))

  def test_activity_dumpsys(self):
    package_name = 'com.world.hello'
    self._mock_execute_command.return_value = b'My awesome dumpsys output!!!'
    activity_dumpsys = self._adb_controller.get_activity_dumpsys(
        package_name, timeout=_TIMEOUT)
    self._mock_execute_command.assert_called_once_with(
        self._adb_controller,
        ['shell', 'dumpsys', 'activity', package_name, package_name], _TIMEOUT)
    # Compare activity_dumpsys to what we want. Notice that we expect a UTF-8
    # string, NOT bytes.
    self.assertEqual(activity_dumpsys, 'My awesome dumpsys output!!!')

  def test_input_tap(self):
    self._mock_execute_command.return_value = b''
    self._adb_controller.input_tap(123, 456, timeout=_TIMEOUT)
    self._mock_execute_command.assert_called_once_with(
        self._adb_controller, ['shell', 'input', 'tap', '123', '456'], _TIMEOUT)

  def test_input_text(self):
    self._mock_execute_command.return_value = b''
    self._adb_controller.input_text('my_text', timeout=_TIMEOUT)
    self._mock_execute_command.assert_called_once_with(
        self._adb_controller, ['shell', 'input', 'text', 'my_text'], _TIMEOUT)

  def test_input_key(self):
    self._mock_execute_command.return_value = b''
    self._adb_controller.input_key('KEYCODE_HOME', timeout=_TIMEOUT)
    self._adb_controller.input_key('KEYCODE_BACK', timeout=_TIMEOUT)
    self._adb_controller.input_key('KEYCODE_ENTER', timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller,
                  ['shell', 'input', 'keyevent', 'KEYCODE_HOME'], _TIMEOUT),
        mock.call(self._adb_controller,
                  ['shell', 'input', 'keyevent', 'KEYCODE_BACK'], _TIMEOUT),
        mock.call(self._adb_controller,
                  ['shell', 'input', 'keyevent', 'KEYCODE_ENTER'], _TIMEOUT),
    ])

    # A key code outside of the accepted codes should raise an exception.
    self.assertRaises(AssertionError, self._adb_controller.input_key,
                      'KEYCODE_0')

  def test_install_apk(self):
    self._mock_execute_command.return_value = b''
    # Passing an invalid path should raise an exception.
    self.assertRaises(AssertionError, self._adb_controller.install_apk, '')

    local_apk_path = os.path.join(absltest.get_default_test_tmpdir(),
                                  'my_app.apk')
    with open(local_apk_path, 'wb') as f:
      f.write(b'blah. whatever')
    self._adb_controller.install_apk(local_apk_path, timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller,
                  ['install', '-r', '-t', '-g', local_apk_path], _TIMEOUT),
    ])

  def test_start_accessibility_service(self):
    self._mock_execute_command.return_value = b''
    self._adb_controller.start_accessibility_service(
        'my.service', timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller, [
            'shell', 'settings', 'put', 'secure',
            'enabled_accessibility_services', 'my.service'
        ], _TIMEOUT)
    ])

  @mock.patch.object(
      adb_controller.AdbController, '_fetch_current_task_id', autospec=True)
  def test_start_screen_pinning_task_not_found(self,
                                               mock_fetch_current_task_id):
    self._mock_execute_command.return_value = b''
    mock_fetch_current_task_id.return_value = -1
    self.assertRaises(
        errors.AdbControllerScreenPinningError,
        self._adb_controller.start_screen_pinning,
        'my.app.CoolActivity', _TIMEOUT)

  @mock.patch.object(
      adb_controller.AdbController, '_fetch_current_task_id', autospec=True)
  def test_start_screen_pinning(self, mock_fetch_current_task_id):
    self._mock_execute_command.return_value = b''
    mock_fetch_current_task_id.return_value = 123
    self._adb_controller.start_screen_pinning(
        'my.app.CoolActivity', timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(self._adb_controller, ['shell', 'am', 'task', 'lock', '123'],
                  _TIMEOUT)
    ])

  def test_fetch_current_task_id(self):
    full_activity_name = (
        'com.google.example.games.nostalgicracer/'
        'com.google.example.games.nostalgicracer.MainActivity')
    bad_task = (
        '  taskId=8: '
        'com.google.android.apps.maps/com.google.android.maps.MapsActivity '
        'bounds=[0,0][480,320] userId=0 visible=true '
        'topActivity=ComponentInfo{%s}' % full_activity_name)
    good_task_not_visible = (
        'taskId=49: %s bounds=[0,0][320,480] userId=0 visible=false more_stuff'
        % full_activity_name)
    good_task = (
        'taskId=50: %s bounds=[0,0][320,480] userId=0 visible=true more_stuff' %
        full_activity_name)
    stack_list = '\n'.join([bad_task, good_task_not_visible,
                            good_task]).encode('utf-8')
    self._mock_execute_command.return_value = stack_list

    self.assertEqual(
        50,
        self._adb_controller._fetch_current_task_id(
            full_activity_name, timeout=_TIMEOUT))

  def test_check_install_not_installed(self):
    self._mock_execute_command.return_value = b"""
package:foo
package:bar
package:baz
"""
    self.assertFalse(
        self._adb_controller.is_package_installed('faz', timeout=_TIMEOUT))

  def test_check_install_installed(self):
    self._mock_execute_command.return_value = b"""
package:foo
package:bar
package:baz
"""
    self.assertTrue(
        self._adb_controller.is_package_installed('baz', timeout=_TIMEOUT))

  def test_uninstall_package_installed(self):
    self._mock_execute_command.return_value = b"""
package:foo
package:bar
package:baz
"""
    self._adb_controller.uninstall_package('foo', timeout=_TIMEOUT)
    self._mock_execute_command.assert_has_calls([
        mock.call(
            self._adb_controller, ['shell', 'pm', 'list', 'packages'],
            timeout=_TIMEOUT),
        mock.call(self._adb_controller, ['uninstall', 'foo'], _TIMEOUT),
    ])

  def test_uninstall_package_not_installed(self):
    self._mock_execute_command.return_value = b"""
package:bar
package:baz
"""
    self._adb_controller.uninstall_package('foo', timeout=_TIMEOUT)
    # Uninstall command not called.
    self._mock_execute_command.assert_called_once_with(
        self._adb_controller, ['shell', 'pm', 'list', 'packages'],
        timeout=_TIMEOUT)

  @parameterized.parameters(
      (True, True, 'null*'),
      (True, False, 'immersive.status=*'),
      (False, True, 'immersive.navigation=*'),
      (False, False, 'immersive.full=*'),
      (None, None, 'immersive.full=*'),  # Defaults to hiding both.
  )
  def test_set_bar_visibility(self, navigation, status, expected):
    expected_output = b'Message.'
    self._mock_execute_command.return_value = expected_output

    self.assertEqual(
        expected_output,
        self._adb_controller.set_bar_visibility(
            navigation=navigation, status=status, timeout=_TIMEOUT))
    self._mock_execute_command.assert_has_calls([
        mock.call(
            self._adb_controller,
            ['shell', 'settings', 'put', 'global', 'policy_control', expected],
            _TIMEOUT),
    ])

  @mock.patch.object(time, 'sleep', autospec=True)
  def test_init_server(self, mock_sleep):
    self._adb_controller.init_server(timeout=_TIMEOUT)
    self._mock_execute_command.assert_called_once_with(self._adb_controller,
                                                       ['devices'], _TIMEOUT)
    mock_sleep.assert_called_once()


class AdbControllerInitTest(absltest.TestCase):

  def test_deletes_problem_env_vars(self):
    os.environ['ANDROID_HOME'] = '/usr/local/Android/Sdk'
    os.environ['ANDROID_ADB_SERVER_PORT'] = '1337'
    adb_controller.AdbController(
        adb_path='my_adb',
        device_name='awesome_device',
        adb_server_port=9999,
        prompt_regex='l33t>',
        default_timeout=_TIMEOUT)
    self.assertNotIn('ANDROID_HOME', os.environ)
    self.assertNotIn('ANDROID_ADB_SERVER_PORT', os.environ)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for adb_log_stream."""

import subprocess

from absl.testing import absltest
from android_env.components import adb_log_stream
import mock


class FakeAdbSubprocess():

  @property
  def stdout(self):
    return [f'line_{i}' for i in range(100)]

  def kill(self):
    pass


class AdbLogStreamTest(absltest.TestCase):

  @mock.patch.object(subprocess, 'Popen', return_value=FakeAdbSubprocess())
  def test_get_stream_output(self, mock_popen):
    stream = adb_log_stream.AdbLogStream(adb_command_prefix=['foo'])
    stream.set_log_filters(['bar'])
    stream_output = stream.get_stream_output()

    for i, line in enumerate(stream_output):
      self.assertEqual(line, f'line_{i}')

    mock_popen.assert_called_with(
        ['foo', 'logcat', '-v', 'epoch', 'bar', '*:S'],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        bufsize=1,
        universal_newlines=True)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract class for handling a stream of logs from a simulator."""

import abc
from typing import List
from absl import logging


class LogStream(metaclass=abc.ABCMeta):
  """Manages the stream of logs output by a simulator."""

  def __init__(self, verbose: bool = False):
    self._verbose = verbose
    self._filters = []

  def get_stream_output(self):
    """Starts log process and returns the stream of logs."""
    for line in self._get_stream_output():
      if self._verbose:
        logging.info('line: %r', line)
      yield line

  @abc.abstractmethod
  def _get_stream_output(self):
    """Starts log process and returns the stream of logs."""
    pass

  @abc.abstractmethod
  def stop_stream(self):
    """Stops the log stream from the simulator."""
    pass

  def set_log_filters(self, log_filters: List[str]):
    """Sets the filters for the log stream."""
    self._filters = list(log_filters) + ['*:S']



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for AndroidEnv."""

from absl.testing import absltest
from android_env import environment
from android_env.components import coordinator as coordinator_lib
import dm_env
import mock
import numpy as np


def _create_mock_env():
  coordinator = mock.create_autospec(coordinator_lib.Coordinator)
  coordinator.action_spec.return_value = {
      'action_type':
          dm_env.specs.DiscreteArray(num_values=3),
      'touch_position':
          dm_env.specs.BoundedArray(
              shape=(2,), dtype=np.float32, minimum=0.0, maximum=1.0),
  }
  coordinator.observation_spec.return_value = {
      'pixels': dm_env.specs.Array(shape=(123, 456, 3), dtype=np.uint8),
      'timedelta': dm_env.specs.Array(shape=(), dtype=np.int64),
      'orientation': dm_env.specs.Array(shape=(4,), dtype=np.uint8),
  }
  coordinator.task_extras_spec.return_value = {
      'click': dm_env.specs.Array(shape=(), dtype=np.int64),
  }
  return environment.AndroidEnv(coordinator)


class AndroidEnvTest(absltest.TestCase):

  def test_specs(self):
    env = _create_mock_env()

    # Check action spec.
    self.assertNotEmpty(env.action_spec())
    self.assertIn('action_type', env.action_spec())
    self.assertIsInstance(env.action_spec()['action_type'],
                          dm_env.specs.DiscreteArray)
    self.assertIn('touch_position', env.action_spec())
    self.assertIsInstance(env.action_spec()['touch_position'],
                          dm_env.specs.BoundedArray)

    # Check observation spec.
    self.assertNotEmpty(env.observation_spec())
    self.assertIn('pixels', env.observation_spec())
    self.assertIsInstance(env.observation_spec()['pixels'], dm_env.specs.Array)
    # The `pixels` entry in the observation spec should match the screen size of
    # the simulator with three color channels (RGB).
    self.assertEqual(env.observation_spec()['pixels'].shape, (123, 456, 3))
    self.assertIn('timedelta', env.observation_spec())
    self.assertIsInstance(env.observation_spec()['timedelta'],
                          dm_env.specs.Array)
    # The `timedelta` should be a scalar.
    self.assertEqual(env.observation_spec()['timedelta'].shape, ())
    self.assertIn('orientation', env.observation_spec())
    # The `orientation` should be a one-hot vector with four dimensions.
    self.assertIsInstance(env.observation_spec()['orientation'],
                          dm_env.specs.Array)
    self.assertEqual(env.observation_spec()['orientation'].shape, (4,))

    # Check extras spec.
    self.assertNotEmpty(env.task_extras_spec())
    self.assertIn('click', env.task_extras_spec())
    self.assertEqual(env.task_extras_spec()['click'].shape, ())
    self.assertEqual(env.task_extras_spec()['click'].dtype, np.int64)

  def test_reset_and_step(self):
    coordinator = mock.create_autospec(coordinator_lib.Coordinator)
    coordinator.action_spec.return_value = {
        'action_type':
            dm_env.specs.DiscreteArray(num_values=3),
        'touch_position':
            dm_env.specs.BoundedArray(
                shape=(2,), dtype=np.float32, minimum=0.0, maximum=1.0),
    }
    coordinator.observation_spec.return_value = {
        'pixels': dm_env.specs.Array(shape=(123, 456, 3), dtype=np.uint8),
        'timedelta': dm_env.specs.Array(shape=(), dtype=np.int64),
        'orientation': dm_env.specs.Array(shape=(4,), dtype=np.uint8),
    }
    coordinator.task_extras_spec.return_value = {
        'click': dm_env.specs.Array(shape=(1,), dtype=np.int64),
    }
    env = environment.AndroidEnv(coordinator)
    coordinator.execute_action.return_value = (
        {
            'pixels': np.random.rand(987, 654, 3),
            'timedelta': 123456,
            'orientation': np.array((1, 0, 0, 0)),
        },  # Observation
        0.0,  # Reward.
        {
            'click': np.array([[246]], dtype=np.int64)
        },  # Task extras.
        False  # Episode ended?
    )

    ts = env.reset()
    self.assertIsInstance(ts, dm_env.TimeStep)
    # After a `reset()` the TimeStep should follow some expectations.
    self.assertTrue(ts.first())
    self.assertEqual(ts.reward, 0.0)
    self.assertEqual(ts.discount, 0.0)
    obs = ts.observation
    self.assertIn('pixels', obs)
    self.assertEqual(obs['pixels'].shape, (987, 654, 3))
    self.assertIn('timedelta', obs)
    self.assertEqual(obs['timedelta'], 123456)
    self.assertIn('orientation', obs)
    self.assertEqual(obs['orientation'].shape, (4,))
    np.testing.assert_equal(obs['orientation'], (1, 0, 0, 0))

    # Extras should also be provided.
    extras = env.task_extras()
    self.assertIn('click', extras)
    self.assertEqual(extras['click'], np.array([246], dtype=np.int64))

    coordinator.get_logs.return_value = {
        'my_measurement': 135,
    }

    # Step again in the environment and check expectations again.
    ts = env.step({'action_type': 1, 'touch_position': (10, 20)})
    self.assertIsInstance(ts, dm_env.TimeStep)
    # The StepType now should NOT be FIRST.
    self.assertFalse(ts.first())
    self.assertEqual(ts.reward, 0.0)
    self.assertEqual(ts.discount, 0.0)
    obs = ts.observation
    self.assertIn('pixels', obs)
    self.assertEqual(obs['pixels'].shape, (987, 654, 3))
    self.assertIn('timedelta', obs)
    self.assertEqual(obs['timedelta'], 123456)
    self.assertIn('orientation', obs)
    self.assertEqual(obs['orientation'].shape, (4,))
    np.testing.assert_equal(obs['orientation'], (1, 0, 0, 0))

    # Extras should still be provided.
    extras = env.task_extras()
    self.assertIn('click', extras)
    self.assertEqual(extras['click'], np.array([246], dtype=np.int64))

    # At this point these methods and properties should return something.
    self.assertNotEmpty(env.android_logs())
    self.assertNotEmpty(env.raw_observation)
    self.assertNotEmpty(env.raw_action)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

import abc
import re
import itertools
import functools
import enum

from typing import Generic, TypeVar, Callable, Optional, Any, Union
from typing import Tuple, List, Pattern
from typing import Iterable

from absl import logging

# The whole data flow for the event flag classes:
# value -> verification -> transform -> wrap -> update
#        \______________/ \__________________________/
#         in EventSource          in EventSlot

I = TypeVar("Input")
V = TypeVar("Verified")
C = TypeVar("Cast")
T = TypeVar("Transformed")
W = TypeVar("Wrapped")

class Event(abc.ABC, Generic[W]):
    #  abstract class `Event` {{{ # 
    #def __init__(self,
            #transformation: Optional[str] = None,
            #cast: Optional[Callable[[V], C]] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None):
        ##  method `__init__` {{{ # 
        #"""
        #transformation - str representing a python expression that can be
          #evaluated by `eval` function or None for a default identity mapping.
          #"x" should be used to indicate the input variable. `x` could be a
          #scalar or a list.
        #cast - callable accepting the verified type returning the cast type or
          #None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        #"""
#
        #self._transformation_str: Optional[str] = transformation
        #self._cast: Callable[[V], C] = cast or (lambda x: x)
        #self._wrap: Callable[[T], W] = wrap or (lambda x: x)
        #update = update or (lambda x, y: y)
        #self._update: Callable[[Optional[W], W], W] = lambda x, y: update(x, y) if x is not None else y
#
        #self._flag_cache: bool = False
        #self._value_cache: Optional[W] = None
#
        #self._ever_set: bool = False
        #self._prerequisites: List[Event] = []
        ##  }}} method `__init__` # 

    #@abc.abstractmethod
    #def set(self, value: I):
        #"""
        #value - the input type
        #"""
#
        #raise NotImplementedError()
    @abc.abstractmethod
    def is_set(self) -> bool:
        """
        return bool
        """

        raise NotImplementedError()
    @abc.abstractmethod
    def get(self) -> List[W]:
        """
        return list of the wrapped type
        """

        raise NotImplementedError()
    @abc.abstractmethod
    def is_ever_set(self) -> bool:
        """
        return bool
        """

        raise NotImplementedError()
    @abc.abstractmethod
    def reset(self):
        raise NotImplementedError()
#
    #@abc.abstractmethod
    #def _clear(self):
        #raise NotImplementedError()
#
    #def get_cache(self) -> Tuple[bool, Optional[W]]:
        ##  method `get_cache` {{{ # 
        #return self._flag_cache, self._value_cache
        ##  }}} method `get_cache` # 
    #def clear_cache(self):
        ##  method `clear_cache` {{{ # 
        #self._flag_cache = False
        #self._value_cache = False
        #self._clear_cache()
        ##  }}} method `clear_cache` # 
    #@abc.abstractmethod
    #def _clear_cache(self):
        #raise NotImplementedError()
    #  }}} abstract class `Event` # 

class EventSource(Event[V], abc.ABC, Generic[I, V]):
    #  abstract class `EventSource` {{{ # 
    class Repeatability(enum.IntEnum):
        NONE = 0
        LAST = 1
        UNLIMITED = 2

    def __init__(self,
            #transformation: Optional[str] = None,
            #cast: Optional[Callable[[V], C]] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: Repeatability = Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        #transformation - str representing a python expression that can be
          #evaluated by `eval` function or None for a default identity mapping.
          #"x" should be used to indicate the input variable. `x` could be a
          #scalar or a list.
        #cast - callable accepting the verified type returning the cast type or
          #None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        super(EventSource, self).__init__()

        self._repeatability: EventSource.Repeatability = repeatability
        self._last_input: I = None # the last input, both activating and not activating
        self._input_history: List[I] = [] # history list of the activating inputs

        self._flag: bool = False
        self._value: List[V] = []

        self._flag_cache: bool = False
        self._value_cache: List[V] = []

        self._ever_set: bool = False
        #  }}} method `__init__` # 

    def set(self, value: I):
        #  method `set` {{{ # 
        """
        value - the input type
        """

        #if any(map(lambda evt: not evt.is_ever_set(), self._prerequisites)):
            #return

        if self._repeatability==EventSource.Repeatability.NONE and value in self._input_history or\
                self._repeatability==EventSource.Repeatability.LAST and value == self._last_input:
            return
        self._last_input = value

        is_validate, value_ = self._verify(value)
        if is_validate:
            self._input_history.append(value)

            #value_ = self._transform(value_)
            #value_ = self._wrap(value_)
            #self._value = self._update(self._value, value_)
            self._value.append(value_)

            self._flag = True
            self._ever_set = True
        #  }}} method `set` # 
    @abc.abstractmethod
    def _verify(self, value: I) -> Tuple[bool, Optional[Union[V, List[V]]]]:
        #  abstract method `_verify` {{{ # 
        """
        This is a hook function for the extenders to implement.

        value - the input type

        return
        - bool
        - the verified type
        """

        raise NotImplementedError()
        #  }}} abstract method `_verify` # 

    def snapshot(self):
        #  method `snapshot` {{{ # 
        self._flag_cache = self._flag
        self._value_cache = self._value
        self._flag = False
        self._value = []
        #  }}} method `snapshot` # 
    def is_set(self) -> bool:
        return self._flag_cache
    def get(self) -> List[W]:
        #  method `get` {{{ # 
        """
        return lsit the wrapped type
        """

        return self._value_cache if self.is_set() else []
        #  }}} method `get` # 
    def clear(self):
        #  method `clear` {{{ # 
        self._flag_cache = False
        self._value_cache = []
        #  }}} method `clear` # 

    def is_ever_set(self) -> bool:
        return self._ever_set
    def reset(self):
        self._ever_set = False
        self._input_history = []
        self._last_input = None
    #  }}} abstract class `EventSource` # 

class EventSlot(Event[W], abc.ABC, Generic[V, T, W]):
    #  abstract class `EventSlot` {{{ # 
    def __init__(self,
            sources: Iterable[Event[V]],
            #cast: Optional[Callable[[V], C]] = None,
            transformation: Optional[List[str]] = None,
            wrap: Optional[Callable[[T], W]] = None,
            update: Optional[Callable[[W, W], W]] = None):
        #  method `__init__` {{{ # 
        """
        sources - iterable of Event
        #cast - callable accepting the verified type returning the cast type or
          #None
        transformation - list of str or None
        wrap - callable accepting the transformed type returning the wrapped
          type or None
        update - callable accepting
          + the wrapped type
          + the wrapped type
          and returning the wrapped type
        """

        super(EventSlot, self).__init__()

        self._sources: List[Event[V]] = list(sources)

        #self._cast: Callable[[V], C] = cast or (lambda x: x)
        self._transformation_str: List[str] = transformation or []
        if len(self._transformation_str)==0:
            self._transformation_str.append("y = x")
        self._wrap: Callable[[T], W] = wrap or (lambda x: x)
        update = update or (lambda x, y: y)
        #self._update: Callable[[Optional[W], W], W] = lambda x, y: update(x, y) if x is not None else y
        self._update: Callable[[W, W], W] = update

        self._ever_set: bool = False
        self._prerequisites: List[Event] = []
        #  }}} method `__init__` # 
    def replace_sources(self, index: int, source: Event):
        #  method `append_sources` {{{ # 
        self._sources[index] = source
        #  }}} method `append_sources` # 

    def _transform(self, x: Union[V, List[V]]) -> T:
        #  method `_transformation` {{{ # 
        """
        x - the same type as `self._value` or list of instances of type of
          `self._value`

        return something
        """

        #try:
            #x = list(map(self._cast, x0)) if hasattr(x0, "__iter__") else self._cast(x0)
        #except:
            #x = x0
        #return eval(self._transformation_str) if self._transformation_str else x
        local_env = {"x": x}
        for cmmd in self._transformation_str:
            #logging.info("\x1b[1;33m{:}\x1b[0m".format(cmmd))
            exec(cmmd, globals(), local_env)
        return local_env["y"]
        #  }}} method `_transformation` # 

    def add_prerequisites(self, *prerequisites: Iterable[Event]):
        #  method `add_prerequisites` {{{ # 
        """
        prerequisites - iterable of Event
        """

        self._prerequisites += prerequisites
        #  }}} method `add_prerequisites` # 
    def _satisfies_prerequisites(self) -> bool:
        return all(map(lambda evt: evt.is_ever_set(), self._prerequisites))

    def is_set(self) -> bool:
        #  method `is_set` {{{ # 
        if not self._satisfies_prerequisites():
            return False
        if self._is_set():
            self._ever_set = True
            return True
        return False
        #  }}} method `is_set` # 
    @abc.abstractclassmethod
    def _is_set(self) -> bool:
        #  method `_is_set` {{{ # 
        raise NotImplementedError()
        #  }}} method `_is_set` # 

    def get(self) -> List[W]:
        #  method `get` {{{ # 
        """
        return list of the wrapped type
        """

        if not self.is_set():
            return []
        return self._get()
        #  }}} method `get` # 
    @abc.abstractclassmethod
    def _get(self) -> List[W]:
        raise NotImplementedError()

    def is_ever_set(self) -> bool:
        if not self._ever_set:
            return self.is_set()
        return True
    def reset(self):
        #  method `reset` {{{ # 
        self._ever_set = False
        for evt in self._sources:
            evt.reset()
        #for evt in self._prerequisites:
            #evt.reset()
        #  }}} method `reset` # 
    #  }}} abstract class `EventSlot` # 

class EmptyEvent(EventSlot[None, None, None]):
    #  class `EmptyEvent` {{{ # 
    def __init__(self):
        super(EmptyEvent, self).__init__([])
    def _is_set(self) -> bool:
        return False
    def _get(self) -> List[None]:
        return []
    #  }}} class `EmptyEvent` # 

class DefaultEvent(EventSlot[V, T, W]):
    #  class `DefaultEvent` {{{ # 
    def __init__(self,
            sources: List[Event[V]],
            #cast: Optional[Callable[[V], C]] = None,
            transformation: Optional[List[str]] = None,
            wrap: Optional[Callable[[T], W]] = None,
            update: Optional[Callable[[W, W], W]] = None):
        #  method `__init__` {{{ # 
        """
        sources - list of Event
        #cast - callable accepting the verified type returning the cast type or
          #None
        transformation - list of str or None
        wrap - callable accepting the transformed type returning the wrapped
          type or None
        update - callable accepting
          + the wrapped type
          + the wrapped type
          and returning the wrapped type
        """

        super(DefaultEvent, self).__init__(sources, transformation, wrap, update)
        #  }}} method `__init__` # 

    def _is_set(self) -> bool:
        #  method `_is_set` {{{ # 
        """
        return bool
        """

        return len(self._sources)>0 and self._sources[0].is_set()
        #  }}} method `_is_set` # 

    def _get(self) -> List[W]:
        source_values = self._sources[0].get()
        return [functools.reduce(self._update,
                    map(self._wrap,
                        map(self._transform,
                            source_values)))]
    #  }}} class `DefaultEvent` # 

class Or(EventSlot[V, T, W]):
    #  class `Or` {{{ # 
    def __init__(self,
            events: Iterable[Event[V]],
            transformation: Optional[List[str]] = None,
            #cast: Optional[Callable[[V], C]] = None,
            wrap: Optional[Callable[[T], W]] = None,
            update: Optional[Callable[[W, W], W]] = None):
        #  method `__init__` {{{ # 
        """
        events - iterable of Event
        transformation - list of str or None
        #cast - callable accepting the verified type returning the cast type or
          #None
        wrap - callable accepting the transformed type returning the wrapped
          type or None
        update - callable accepting
          + the wrapped type
          + the wrapped type
          and returning the wrapped type
        """

        #update = update or (lambda x, y: x)
        super(Or, self).__init__(events, transformation, wrap, update)

        #self._events: List[Event[Any, Any, Any, Any, V]] = list(events)
        #  }}} method `__init__` # 

    def _is_set(self) -> bool:
        #  method `_is_set` {{{ # 
        """
        return bool
        """

        return any(map(lambda x: x.is_set(), self._sources))
        #  }}} method `_is_set` # 

    def _get(self) -> List[W]:
        #  method `_get` {{{ # 
        """
        return list of the wrapped type
        """

        return [functools.reduce(self._update,
                    map(self._wrap,
                        map(self._transform,
                            itertools.chain.from_iterable(
                                map(lambda evt: evt.get(),
                                    self._sources)))))]
        #value = None
        #is_set = False
        #for evt in self._events:
            #if evt.is_set():
                #is_set = True
                #self._ever_set = True
                #value = self._update(value, self._wrap(self._transform(evt.get())))
        #return value if is_set else None
        #  }}} method `_get` # 
    #  }}} class `Or` # 

class And(EventSlot[V, T, W]):
    #  class `And` {{{ # 
    def __init__(self,
            events: Iterable[Event[V]],
            transformation: Optional[List[str]] = None,
            #cast: Optional[Callable[[V], C]] = None,
            wrap: Optional[Callable[[T], W]] = None):
            #update: Optional[Callable[[W, W], W]] = None):
        #  method `__init__` {{{ # 
        """
        events - iterable of Event
        transformation - list of str of None
        #cast - callable accepting the verified type returning the cast type or
          #None
        wrap - callable accepting the transformed type returning the wrapped
          type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        """

        super(And, self).__init__(events, transformation, wrap)

        #self._events: List[Event[Any, Any, Any, Any, V]] = list(events)
        #  }}} method `__init__` # 

    def _is_set(self) -> bool:
        #  method `_is_set` {{{ # 
        """
        return bool
        """

        if len(self._sources)==0:
            return False
        return all(map(lambda x: x.is_set(), self._sources))
        #  }}} method `_is_set` # 

    def _get(self) -> List[W]:
        #  method `_get` {{{ # 
        """
        return list of the wrapped type
        """

        return [self._wrap(
                    self._transform(
                        list(
                            map(lambda x: x.get(),
                                self._sources))))]
        #  }}} method `_get` # 
    #  }}} class `And` # 

class RegionEvent(EventSource[I, V], abc.ABC):
    #  abstract class `RegionEvent` {{{ # 
    def __init__(self, region: Iterable[float], needs_detection: bool = False,
            #transformation: Optional[str] = None,
            #cast: Optional[Callable[[V], C]] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        region - iterable of four floats as [x0, y0, x1, y1]
        needs_detection - bool
        #transformation - str of None
        #cast - callable accepting the verified type returning the cast type or
          #None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        #super(RegionEvent, self).__init__(transformation, cast, wrap, update, repeatability)
        super(RegionEvent, self).__init__(repeatability)
        self._region: List[float] = list(itertools.islice(region, 4))
        self._needs_detection: bool = needs_detection
        #  }}} method `__init__` # 

    @property
    def region(self) -> List[float]:
        return self._region
    @property
    def needs_detection(self) -> bool:
        return self._needs_detection
    #  }}} abstract class `RegionEvent` # 

class TextEvent(RegionEvent[Optional[str], str]):
    #  class `TextEvent` {{{ # 
    def __init__(self, expect: str,
            region: Iterable[float], needs_detection: bool = False,
            #transformation: Optional[str] = None,
            #cast: Optional[Callable[[str], C]] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        expect - str
        region - iterable of four floats as [x0, y0, x1, y1]
        needs_detection - bool
        #transformation - str or None
        #cast - callable accepting the verified type returning the cast type or
          #None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        #super(TextEvent, self).__init__(region, needs_detection, transformation, cast, wrap, update, repeatability)
        super(TextEvent, self).__init__(region, needs_detection, repeatability)

        self._expect: Pattern[str] = re.compile(expect)
        #  }}} method `__init__` # 

    def _verify(self, text: Optional[str]) -> Tuple[bool, Optional[List[str]]]:
        #  method `_verify` {{{ # 
        """
        text - str or None

        return bool, list of str or None
        """

        if text is None:
            return False, None
        match_ = self._expect.search(text)
        if match_ is not None:
            return True, match_.groups()
        return False, None
        #  }}} method `_verify` # 
    #  }}} class `TextEvent` # 

class IconRecogEvent(RegionEvent[str, bool]):
    #  class `IconRecogEvent` {{{ # 
    def __init__(self, icon_class: str,
            region: Iterable[float], needs_detection: bool = False,
            #transformation: Optional[str] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        icon_class - str
        region - iterable of four floats as [x0, y0, x1, y1]
        needs_detection - bool
        #transformation - str of None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        #super(IconRecogEvent, self).__init__(region, needs_detection, transformation, None, wrap, update, repeatability)
        super(IconRecogEvent, self).__init__(region, needs_detection, repeatability)

        self._icon_class: str = icon_class
        #  }}} method `__init__` # 

    def _verify(self, icon_class: str) -> Tuple[bool, bool]:
        #  method `_verify` {{{ # 
        """
        icon_class - str

        return bool, bool
        """

        flag = icon_class==self._icon_class
        return flag, flag
        #  }}} method `_verify` # 
    #  }}} class `IconRecogEvent` # 

class IconMatchEvent(RegionEvent[bool, bool]):
    #  class `IconMatchEvent` {{{ # 
    def __init__(self, path,
            region: Iterable[float], needs_detection: bool = False,
            #transformation: Optional[str] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        path - str
        region - iterable of four floats as [x0, y0, x1, y1]
        needs_detection - bool
        #transformation - str or None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        #super(IconMatchEvent, self).__init__(region, needs_detection, transformation, None, wrap, update, repeatability)
        super(IconMatchEvent, self).__init__(region, needs_detection, repeatability)

        self._path: str = path
        #  }}} method `__init__` # 

    @property
    def path(self) -> str:
        return self._path

    def _verify(self, set_flag: bool) -> Tuple[bool, bool]:
        #  method `_verify` {{{ # 
        """
        set_flag - bool

        return bool, bool
        """

        return set_flag, set_flag
        #  }}} method `_verify` # 
    #  }}} class `IconMatchEvent` # 

P = TypeVar("Property")

class ViewHierarchyEvent(EventSource[List, Any]):
    #  class `ViewHierarchyEvent` {{{ # 
    class Property(abc.ABC, Generic[P]):
        #  class `Property` {{{ # 
        def __init__(self, name: str):
            #  method `__init__` {{{ # 
            """
            name - str
            """

            self._name: str = name
            #  }}} method `__init__` # 

        @property
        def name(self) -> str:
            return self._name

        @abc.abstractmethod
        def match(self, value: P) -> bool:
            #  abstract method `match` {{{ # 
            """
            value - something

            return bool
            """

            raise NotImplementedError()
            #  }}} abstract method `match` # 
        #  }}} class `Property` # 

    class PureProperty(Property[Any]):
        #  class `PureProperty` {{{ # 
        def __init__(self, name: str):
            super(ViewHierarchyEvent.PureProperty, self).__init__(name)

        def match(self, value: Any) -> bool:
            return True
        #  }}} class `PureProperty` # 

    class ScalarProperty(Property[P]):
        #  class `ScalarProperty` {{{ # 
        def __init__(self, name: str,
                comparator: Callable[[P, P], bool], value: P):
            #  method `__init__` {{{ # 
            """
            name - str
            comparator - callable accepting
              + the same type with `value`
              + the same type with `value`
              and returning bool
            value - something
            """

            super(ViewHierarchyEvent.ScalarProperty, self).__init__(name)
            self._comparator: Callable[[P, P], bool] = comparator
            self._value: P = value
            #  }}} method `__init__` # 

        def match(self, value: P) -> bool:
            #  method `match` {{{ # 
            """
            value - something

            return bool
            """

            return self._comparator(self._value, value)
            #  }}} method `match` # 
        #  }}} class `ScalarProperty` # 

    class StringProperty(Property[str]):
        #  class `StringProperty` {{{ # 
        def __init__(self, name: str, pattern: str):
            #  method `__init__` {{{ # 
            """
            name - str
            pattern - str
            """

            super(ViewHierarchyEvent.StringProperty, self).__init__(name)
            self._pattern: Pattern[str] = re.compile(pattern)
            #  }}} method `__init__` # 

        def match(self, value: str) -> bool:
            #  method `match` {{{ # 
            """
            value - str

            return bool
            """

            return self._pattern.search(value) is not None
            #  }}} method `match` # 
        #  }}} class `StringProperty` # 

    def __init__(self, vh_path: Iterable[str], vh_properties: Iterable[Property],
            #transformation: Optional[str] = None,
            #cast: Optional[Callable[[V], C]] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        vh_path - iterable of str
        vh_property - iterable of Property
        #transformation - str or None
        #cast - callable accepting the verified type returning the cast type or
          #None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        super(ViewHierarchyEvent, self).__init__(repeatability)

        self._vh_path: List[str] = list(vh_path)
        self._vh_properties: List[ViewHierarchyEvent.Property] = list(vh_properties)
        self._property_names: List[str] = list(map(lambda prpt: prpt.name, vh_properties))
        #  }}} method `__init__` # 

    @property
    def path(self) -> List[str]:
        return self._vh_path
    @property
    def property_names(self) -> List[str]:
        return self._property_names

    def _verify(self, values: List[Any]) -> Tuple[bool, Optional[List[Any]]]:
        #  method `_verify` {{{ # 
        """
        values - list with the same length as `self._vh_properties`

        return
        - bool
        - the same type as `values` or None
        """

        if all(map(lambda p: p[0].match(p[1]),
                itertools.zip_longest(self._vh_properties, values))):
            logging.info("\x1b[5;31mVH Event\x1b[0m")
            return True, values
        return False, None
        #  }}} method `_verify` # 
    #  }}} class `ViewHierarchyEvent` # 

class LogEvent(EventSource[str, str]):
    #  class `LogEvent` {{{ # 
    def __init__(self, filters: Iterable[str], pattern: str,
            #transformation: Optional[str] = None,
            #cast: Optional[Callable[[V], C]] = None,
            #wrap: Optional[Callable[[T], W]] = None,
            #update: Optional[Callable[[W, W], W]] = None,
            repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE):
        #  method `__init__` {{{ # 
        """
        filters - iterable of str
        pattern - str
        #cast - callable accepting the verified type returning the cast type or
          #None
        #wrap - callable accepting the transformed type returning the wrapped
          #type or None
        #update - callable accepting
          #+ the wrapped type
          #+ the wrapped type
          #and returning the wrapped type
        repeatability - EventSource.Repeatability
        """

        super(LogEvent, self).__init__(repeatability)

        self._filters: List[str] = list(filters)
        self._pattern: Pattern[str] = re.compile(pattern)
        #  }}} method `__init__` # 

    @property
    def filters(self) -> List[str]:
        return self._filters
    @property
    def pattern(self) -> Pattern[str]:
        return self._pattern

    def _verify(self, line: str) -> Tuple[bool, Optional[List[str]]]:
        #  method `_verify` {{{ # 
        """
        line - str

        return bool, list of str or None
        """

        match_ = self._pattern.search(line)
        if match_ is not None:
            logging.info("\x1b[5;31mLog Event\x1b[0m")
            return True, match_.groups()
        return False, None
        #  }}} method `_verify` # 
    #  }}} class `LogEvent` # 

class ResponseEvent(EventSource[str, str]):
    #  class ResponseEvent {{{ # 
    def __init__( self
                , pattern: str
                , repeatability: EventSource.Repeatability = EventSource.Repeatability.NONE
                ):
        #  method __init__ {{{ # 
        self._pattern: Pattern[str] = re.compile(pattern)
        #  }}} method __init__ # 

    def _verify(self, response: str) -> Tuple[bool, Optional[List[str]]]:
        #  method _verify {{{ # 
        match_ = self._pattern.search(response)
        if match_ is not None:
            logging.info("\x1b[5;31mResponse Event\x1b[0m")
            return True, match_.groups()
        return False, None
        #  }}} method _verify # 
    #  }}} class ResponseEvent # 



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A base class for talking to different types of Android simulators."""

import abc
from typing import Optional, List, Dict, Tuple

from absl import logging
from android_env.components import action_type as action_type_lib
from android_env.components import adb_controller
from android_env.components import log_stream
from android_env.components import utils

import numpy as np


class BaseSimulator(metaclass=abc.ABCMeta):
  """An interface for communicating with an Android simulator."""

  def __init__(self,
               num_fingers: int = 1,
               show_touches: bool = True,
               show_pointer_location: bool = True,
               show_status_bar: bool = False,
               show_navigation_bar: bool = False,
               verbose_logs: bool = False):
    """Instantiates a BaseSimulator object.

    The simulator may be an emulator, virtual machine or even a physical device.
    Each simulator has its own AdbController that is used for internal
    bookkeeping.

    Args:
      num_fingers: Number of virtual fingers of the agent.
      show_touches: Whether to show circles on the screen indicating the
        position of the current touch.
      show_pointer_location: Whether to show blue lines on the screen
        indicating the position of the current touch.
      show_status_bar: Whether or not to show the status bar (at the top of the
        screen, displays battery life, time, notifications etc.).
      show_navigation_bar: Whether or not to show the navigation bar (at the
        bottom of the screen, displayes BACK and HOME buttons, etc.)
      verbose_logs: If true, the log stream of the simulator will be verbose.
    """

    self._num_fingers = num_fingers
    self._show_touches = show_touches
    self._show_pointer_location = show_pointer_location
    self._show_status_bar = show_status_bar
    self._show_navigation_bar = show_navigation_bar
    self._verbose_logs = verbose_logs

    self._orientation = np.zeros(4, dtype=np.uint8)
    self._screen_dimensions = None
    self._last_obs_timestamp = 0
    self._launched = False
    self._adb_controller = None
    self._log_stream = None

  def num_fingers(self) -> int:
    return self._num_fingers

  def is_launched(self) -> bool:
    return self._launched

  @abc.abstractmethod
  def adb_device_name(self) -> str:
    """Returns the device name that the adb client will connect to."""
    pass

  @abc.abstractmethod
  def create_adb_controller(self) -> adb_controller.AdbController:
    """Returns an ADB controller which can communicate with this simulator."""
    pass

  @abc.abstractmethod
  def _restart_impl(self) -> None:
    """Platform specific restart implementation."""
    pass

  @abc.abstractmethod
  def _launch_impl(self) -> None:
    """Platform specific launch implementation."""
    pass

  @abc.abstractmethod
  def send_action(self, action: Dict[str, np.ndarray]) -> None:
    """Sends the action to be executed to the simulator."""
    pass

  @abc.abstractmethod
  def _get_observation(self) -> Optional[List[np.ndarray]]:
    """Implementation of the Android observation.

    We expect the following items to be present in the output:
      [0] (image): Numpy array. The screenshot of the simulator. The numpy array
          has shape [height, width, num_channels] and can be loaded into PIL
          using Image.fromarray(img, mode='RGB') and be saved as a PNG file
          using my_pil.save('/tmp/my_screenshot.png', 'PNG').
      [1] (timedelta): np.int64 The number of microseconds since Unix Epoch.
    """
    pass

  @abc.abstractmethod
  def _create_log_stream(self) -> log_stream.LogStream:
    """Creates a stream of logs from the simulator."""
    pass

  def launch(self) -> None:
    """Launches the simulator."""
    if not self._launched:
      self._launched = True
      self._launch_impl()
      self._post_launch_setup()
    else:
      self.restart()

  def restart(self) -> None:
    """Restarts the simulator."""
    logging.info('Restarting the simulator...')
    self._adb_controller.close()
    self._restart_impl()
    self._post_launch_setup()
    logging.info('Done restarting the simulator.')

  def _post_launch_setup(self) -> None:
    """Performs necessary steps after the simulator has been launched."""
    self._screen_dimensions = np.array(
        self._adb_controller.get_screen_dimensions())
    self._adb_controller.set_touch_indicators(
        show_touches=self._show_touches,
        pointer_location=self._show_pointer_location)
    self._adb_controller.set_bar_visibility(
        navigation=self._show_navigation_bar,
        status=self._show_status_bar)
    self._log_stream = self._create_log_stream()

  def screen_dimensions(self) -> np.ndarray:
    """Returns the screen dimensions in pixels.

    IMPORTANT: This value is only valid after a successful launch() call.
    """
    if self._screen_dimensions is None:
      raise AssertionError('Screen dimension not available yet.')
    return self._screen_dimensions

  def update_device_orientation(self) -> None:
    """Updates the current device orientation."""

    # Skip fetching the orientation if we already have it.
    if not np.all(self._orientation == np.zeros(4)):
      logging.info('self._orientation already set, not setting it again')
      return

    orientation = self._adb_controller.get_orientation()
    if orientation not in {'0', '1', '2', '3'}:
      logging.error('Got bad orientation: %r', orientation)
      return

    # Transform into one-hot format.
    orientation_onehot = np.zeros([4], dtype=np.uint8)
    orientation_onehot[int(orientation)] = 1
    self._orientation = orientation_onehot

  def _prepare_action(
      self, action: Dict[str, np.ndarray]
  ) -> Tuple[int, int, bool]:
    """Turns an AndroidEnv action into values that the simulator can interpret.

    Converts float-valued 'touch_position' to integer coordinates corresponding
    to specific pixels, and 'action_type' to booleans indicating whether the
    screen is touched at said location or not. The result of this function can
    be sent directly to the underlying simulator (e.g. the Android Emulator,
    virtual machine, or a phone).

    Args:
      action: An action containing 'action_type' and 'touch_position'.
    Returns:
      A tuple with the format (x: int, y: int, down/up: bool).
    """
    is_touch = action['action_type'].item() == action_type_lib.ActionType.TOUCH
    touch_position = action['touch_position']
    touch_pixels = utils.touch_position_to_pixel_position(
        touch_position, width_height=self._screen_dimensions[::-1])
    return (touch_pixels[0], touch_pixels[1], is_touch)

  def _split_action(
      self, action: Dict[str, np.ndarray]) -> List[Dict[str, np.ndarray]]:
    """Splits a multitouch action into a list of single-touch actions."""

    single_touch_actions = [{
        'action_type': action['action_type'],
        'touch_position': action['touch_position'],
    }]
    for i in range(2, self._num_fingers + 1):
      single_touch_actions.append({
          'action_type': action[f'action_type_{i}'],
          'touch_position': action[f'touch_position_{i}'],
      })
    return single_touch_actions

  def get_observation(self) -> Dict[str, np.ndarray]:
    """Returns the environment observation.

    It calls the simulator implementation of _get_observation() and assembles
    the observation to be in the format expected by AndroidEnv, including
    passing a timestamp delta since the last observation (as opposed to the
    absolute timestamp) and appending the one-hot representation of the device
    orientation.
    """
    pixels, timestamp = self._get_observation()
    timestamp_delta = timestamp - self._last_obs_timestamp
    self._last_obs_timestamp = timestamp
    return {
        'pixels': pixels,
        'timedelta': timestamp_delta,
        'orientation': self._orientation
    }

  def close(self):
    self._adb_controller.close()

  def get_log_stream(self):
    return self._log_stream



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

from android_env.components import thread_function
import enum

#from android_env.proto import emulator_controller_pb2
#from android_env.proto import emulator_controller_pb2_grpc
import numpy as np
import torch
#import torchvision.ops as tvops
import torchvision.io as tvio

from absl import logging
import traceback

from typing import Callable
from typing import List, Tuple
from android_env.components import event_listeners
import threading

# mainly taken from `DumpsysThread`
class ScreenAnalyzerThread(thread_function.ThreadFunction):
    #  class `ScreenAnalyzerThread` {{{ # 
    class Signal(enum.IntEnum):
        #  enum `Signal` {{{ # 
        CHECK_SCREEN = 0
        DID_NOT_CHECK = 1
        CHECK_ERROR = 2
        OK = 4
        #  }}} enum `Signal` # 

    def __init__(self,
            text_detector: Callable[[torch.Tensor, List[torch.Tensor]],
                List[List[str]]],
            text_recognizer: Callable[[torch.Tensor, List[torch.Tensor]],
                List[str]],
            icon_detector: Callable[[torch.Tensor, List[torch.Tensor]],
                Tuple[torch.Tensor, List[List[str]]]],
            icon_recognizer: Callable[[torch.Tensor, List[torch.Tensor]],
                List[str]],
            icon_matcher: Callable[[torch.Tensor, List[torch.Tensor], torch.Tensor],
                List[List[bool]]],
            #emulator_stub, image_format,
            #check_frequency,
            #max_failed_current_activity,
            lock: threading.Lock,
            block_input: bool, block_output: bool, name: str = "screen_analyzer"):
        #  method `__init__` {{{ # 
        """
        text_detector - callable accepting
          + tensor of float32 with shape (3, height, width) as screenshot
          + list of tensor of float32 with shape (1, 4) with length nb_bboxes
          and returning list of [[list of str]] with length nb_bboxes as
          detected texts
        text_recognizer - callable accepting
          + tensor of float32 with shape (3, height, width)
          + list of tensor of float32 with shape (1, 4) with length nb_bboxes
          and returning list of str with length nb_bboxes

        icon_detector - callable accepting
          + tensor of float32 with shape (3, height, width)
          + list of tensor of float32 with shape (1, 4) with length nb_bboxes
          and returning
          + tensor of float32 with shape (nb_bboxes, nb_candidates, 4)
          + list of [[list of str with length nb_candidates]] with length
            nb_bboxes as recognized icon classes
        icon_recognizer - callable accepting
          + tensor of float32 with shape (3, height, width)
          + list of tensor of float32 with shape (1, 4) with length nb_bboxes
          and returning list of str with length nb_bboxes as recognized icon
          class
        icon_matcher - callable accepting
          + tensor of float32 with shape (3, height, width) as the screenshot
          + list of tensor of float32 with shape (3, height', width') with length
            nb_bboxes as the targets
          + tensor of float32 with shape (nb_bboxes, nb_candidates, 4)
          and returning list of [[list of bool with length nb_candidates]] with
          length nb_bboxes

        #emulator_stub - emulator_controller_pb2_grpc.EmulatorControllerStub
        #image_format - emulator_controller_pb2.ImageFormat

        #check_frequency - int
        #max_failed_current_activity - int

        lock - threading.Lock

        block_input - bool
        block_output - bool
        name - str
        """

        self._text_detector: Callable[[torch.Tensor, List[torch.Tensor]],
                List[List[str]]] = text_detector
        self._text_recognizer: Callable[[torch.Tensor, List[torch.Tensor]],
                List[str]] = text_recognizer
        self._icon_detector: Callable[[torch.Tensor, List[torch.Tensor]],
                Tuple[torch.Tensor, List[List[str]]]] = icon_detector
        self._icon_recognizer: Callable[[torch.Tensor, List[torch.Tensor]],
                List[str]] = icon_recognizer
        self._icon_matcher: Callable[[torch.Tensor, List[torch.Tensor], torch.Tensor],
                List[List[bool]]] = icon_matcher

        #self._emulator_stub = emulator_stub
        #self._image_format = image_format

        self._text_recog_event_listeners: List[event_listeners.Event] = []
        self._text_detect_event_listeners: List[event_listeners.Event] = []
        self._icon_recog_event_listeners: List[event_listeners.Event] = []
        self._icon_detect_event_listeners: List[event_listeners.Event] = []
        self._icon_match_event_listeners: List[event_listeners.Event] = []
        self._icon_detect_match_event_listeners: List[event_listeners.Event] = []

        self._lock: threading.Lock = lock

        #self._main_loop_counter = 0
        #self._check_frequency = check_frequency
        #self._max_failed_activity_extraction = max_failed_current_activity
        #self._num_failed_activity_extraction = 0

        super(ScreenAnalyzerThread, self).__init__(block_input, block_output, name)
        #  }}} method `__init__` # 

    #  Methods to Add Event Listeners {{{ # 
    def add_text_event_listeners(self, *event_listeners):
        """
        event_listeners - list of event_listeners.TextEvent
        """

        for lstn in event_listeners:
            if lstn.needs_detection:
                self._text_detect_event_listeners.append(lstn)
            else:
                self._text_recog_event_listeners.append(lstn)

    def add_icon_event_listeners(self, *event_listeners):
        """
        event_listeners - list of event_listeners.IconRecogEvent
        """

        for lstn in event_listeners:
            if lstn.needs_detection:
                self._icon_detect_event_listeners.append(lstn)
            else:
                self._icon_recog_event_listeners.append(lstn)

    def add_icon_match_event_listeners(self, *event_listeners):
        """
        event_listeners - list of event_listeners.IconMatchEvent
        """

        for lstn in event_listeners:
            if lstn.needs_detection:
                self._icon_detect_match_event_listeners.append(lstn)
            else:
                self._icon_match_event_listeners.append(lstn)
    #  }}} Methods to Add Event Listeners # 

    #def get_screenshot(self):
        ##  method `get_screenshot` {{{ # 
        #"""
        #return tensor of float32 with shape (3, height, width)
        #"""
#
        #image_proto = self._emulator_stub.getScreenshot(self._image_format)
        #height = image_proto.format.height
        #width = image_proto.format.width
        #image = np.frombuffer(image_proto.image, dtype=np.uint8, count=height*width*3)
        #image.shape = (height, width, 3)
        #image = np.transpose(image, axes=(2, 0, 1))
        #return torch.tensor(image, dtype=torch.float32)
        ##  }}} method `get_screenshot` # 

    #  Methods to Match Registered Events {{{ # 
    @torch.no_grad()
    def match_text_events(self, screen):
        #  method `match_text_events` {{{ # 
        """
        screen - tensor of float32 with shape (3, height, width)
        """

        _, height, width = screen.shape

        # first, for detection
        bboxes = []
        for lstn in self._text_detect_event_listeners:
            x0, y0, x1, y1 = lstn.region
            bboxes.append(torch.tensor(
                [
                    x0*width,
                    y0*height,
                    x1*width,
                    y1*height
                ])[None, :])
        if len(bboxes)>0:
            results = self._text_detector(screen, bboxes)
            with self._lock:
                for lstn, rslts in zip(self._text_detect_event_listeners, results):
                    for cddt in rslts:
                        lstn.set(cddt)
                        if lstn.is_set():
                            break

        # then, for pure recognize
        try:
            bboxes = []
            for lstn in self._text_recog_event_listeners:
                x0, y0, x1, y1 = lstn.region
                bboxes.append(torch.tensor(
                    [
                        x0*width,
                        y0*height,
                        x1*width,
                        y1*height
                    ])[None, :])
            if len(bboxes)>0:
                results = self._text_recognizer(screen, bboxes)
                with self._lock:
                    for lstn, rslt in zip(self._text_recog_event_listeners, results):
                        #logging.info("ZDY_DEBUG: {:}".format(type(rslt)))
                        lstn.set(rslt)
        except Exception as e:
            logging.error("Error occurred during text recognition!")
            logging.error(str(e))
            traceback.print_exc()
        #  }}} method `match_text_events` # 

    @torch.no_grad()
    def match_icon_events(self, screen):
        #  method `match_icon_events` {{{ # 
        """
        screen - tensor of float32 with shape (3, height, width)
        """

        _, height, width = screen.shape

        # first, for detection
        try:
            bboxes = []
            for lstn in self._icon_detect_event_listeners:
                x0, y0, x1, y1 = lstn.region
                bboxes.append(torch.tensor(
                    [
                        x0*width,
                        y0*height,
                        x1*width,
                        y1*height
                    ])[None, :])
            if len(bboxes)>0:
                _, results = self._icon_detector(screen, bboxes)
                with self._lock:
                    for lstn, rslts in zip(self._icon_detect_event_listeners, results):
                        for cddt in rslts:
                            lstn.set(cddt)
                            if lstn.is_set():
                                #logging.info("Icon Event Set!")
                                break
        except Exception as e:
            logging.error("Error occurred during icon detection!")
            logging.error(str(e))

        # then, for pure recognize
        try:
            bboxes = []
            for lstn in self._icon_recog_event_listeners:
                x0, y0, x1, y1 = lstn.region
                bboxes.append(torch.tensor(
                    [
                        x0*width,
                        y0*height,
                        x1*width,
                        y1*height
                    ])[None, :])
            if len(bboxes)>0:
                results = self._icon_recognizer(screen, bboxes)
                with self._lock:
                    for lstn, rslt in zip(self._icon_recog_event_listeners, results):
                        lstn.set(rslt)
        except Exception as e:
            logging.error("Error occurred during icon recognition!")
            logging.error(str(e))
        #  }}} method `match_icon_events` # 

    @torch.no_grad()
    def match_icon_match_events(self, screen):
        #  method `match_icon_match_events` {{{ # 
        """
        screen - tensor of float32 with shape (3, height, width)
        """

        _, height, width = screen.shape

        # first, for detection
        bboxes = []
        target_images = []
        for lstn in self._icon_detect_match_event_listeners:
            x0, y0, x1, y1 = lstn.region
            bboxes.append(torch.tensor(
                [
                    x0*width,
                    y0*height,
                    x1*width,
                    y1*height
                ])[None, :])

            image_tensor = tvio.read_image(lstn.path, tvio.ImageReadMode.RGB)
            image_tensor = image_tensor.to(dtype=torch.float32)/255.
            target_images.append(image_tensor)
        if len(bboxes)>0:
            candidate_bboxes, _ = self._icon_detector(screen, bboxes)
            base_bboxes = torch.cat(bboxes)[:, None, :]
            candidate_bboxes[:, :, 0] += base_bboxes[:, :, 0]
            candidate_bboxes[:, :, 2] += base_bboxes[:, :, 0]
            candidate_bboxes[:, :, 1] += base_bboxes[:, :, 1]
            candidate_bboxes[:, :, 3] += base_bboxes[:, :, 1]
            results = self._icon_matcher(screen, target_images, candidate_bboxes)

            with self._lock:
                for lstn, rslts in zip(self._icon_detect_match_event_listeners, results):
                    if any(rslts):
                        lstn.set(True)

        # then, for pure recognize
        bboxes = []
        target_images = []
        for lstn in self._icon_match_event_listeners:
            x0, y0, x1, y1 = lstn.region
            bboxes.append(torch.tensor(
                [
                    x0*width,
                    y0*height,
                    x1*width,
                    y1*height
                ])[None, :])

            image_tensor = tvio.read_image(lstn.path, tvio.ImageReadMode.RGB)
            image_tensor = image_tensor.to(dtype=torch.float32)/255.
            target_images.append(image_tensor)
        if len(bboxes)>0:
            bboxes = torch.cat(bboxes)[:, None, :]
            results = self._icon_matcher(screen, target_images, bboxes)
            with self._lock:
                for lstn, rslt in zip(self._icon_recog_event_listeners, results):
                    lstn.set(rslt[0])
        #  }}} method `match_icon_match_events` # 
    #  }}} Methods to Match Registered Events # 

    def main(self):
        #  method `main` {{{ # 
        command = self._read_value()
        if command!=ScreenAnalyzerThread.Signal.CHECK_SCREEN:
            self._write_value(ScreenAnalyzerThread.Signal.DID_NOT_CHECK)
            return

        screen = self._read_value()
        if not isinstance(screen, np.ndarray):
            self._write_value(ScreenAnalyzerThread.Signal.DID_NOT_CHECK)
            return
        screen = np.clip(screen, 0, 255).astype(np.float32)/255.
        screen = torch.from_numpy(screen)

        #self._main_loop_counter += 1
        #if self._check_frequency<=0 or\
                #self._main_loop_counter < self._check_frequency:
            #self._write_value(ScreenAnalyzerThread.Signal.DID_NOT_CHECK)
            #return
        #self._main_loop_counter = 0


        try:
            #logging.info("Screen Analyzer Tries to Catch Screenshot...")
            #screen = self.get_screenshot()
            #logging.info("Screen Analyzer Caught Screenshot!")

            self.match_text_events(screen)
            self.match_icon_events(screen)
            self.match_icon_match_events(screen)

            self._write_value(ScreenAnalyzerThread.Signal.OK)
        except:
            self._write_value(ScreenAnalyzerThread.Signal.CHECK_ERROR)
        #  }}} method `main` # 
    #  }}} class `ScreenAnalyzerThread` # 



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for AndroidEnv wrappers."""

from typing import Any, Dict
import dm_env
from dm_env import specs


class BaseWrapper(dm_env.Environment):
  """AndroidEnv wrapper."""

  def __init__(self, env):
    self._env = env

  def switch_task(self, index: int) -> dm_env.TimeStep:
    self._reset_state()
    timestep = self._env.switch_task(index)
    self._post_switch_task()
    timestep = self._process_timestep(timestep)
    return timestep

  def reset(self) -> dm_env.TimeStep:
    self._reset_state()
    timestep = self._process_timestep(self._env.reset())
    return timestep

  def step(self, action: Any) -> dm_env.TimeStep:
    action = self._process_action(action)
    return self._process_timestep(self._env.step(action))

  def _post_switch_task(self):
    pass

  def _reset_state(self):
    pass

  def _process_action(self, action: Any) -> Any:
    return action

  def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
    return timestep

  def observation_spec(self) -> Dict[str, specs.Array]:
    return self._env.observation_spec()

  def action_spec(self) -> Dict[str, specs.Array]:
    return self._env.action_spec()

  def _wrapper_logs(self) -> Dict[str, Any]:
    """Add wrapper specific logging here."""
    return {}

  def android_logs(self) -> Dict[str, Any]:
    info = self._env.android_logs()
    info.update(self._wrapper_logs())
    return info

  @property
  def raw_env(self):
    """Recursively unwrap until we reach the true 'raw' env."""
    wrapped = self._env
    if hasattr(wrapped, 'raw_env'):
      return wrapped.raw_env
    return wrapped

  def __getattr__(self, attr):
    """Delegate attribute access to underlying environment."""
    return getattr(self._env, attr)

  def close(self):
    self._env.close()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A utility class to run a function in a separate daemon thread."""

import abc
import enum
import queue
import threading
from typing import Optional

from absl import logging


class ThreadFunction(metaclass=abc.ABCMeta):
  """Base class that encapsulates long-lived functions in a separate thread."""

  class Signal(enum.IntEnum):
    """Defines commands we can use to communicate with the internal thread."""
    # The thread should be stopped to allow for graceful termination.
    KILL = 1

  def __init__(self, block_input: bool, block_output: bool, name: str):
    """Initializes this ThreadFunction.

    Args:
      block_input: Whether to block this thread when reading its input queue.
      block_output: Whether to block this thread when writing to its
          output queue.
      name: Name of the thread. Used to keep track of threads in logging.
    """
    self._block_input = block_input
    self._block_output = block_output
    self._name = name

    self._input_queue = queue.Queue()
    self._output_queue = queue.Queue()
    self._should_run = True
    self._internal_thread = threading.Thread(target=self._run)
    self._internal_thread.daemon = True
    self._internal_thread.start()

  def read(self, block: bool = True, timeout: Optional[float] = None):
    """'Public' method for clients to read values _from_ this thread.

    Args:
      block: Whether the client should block.
      timeout: Timeout for getting output from the queue, in seconds.
    Returns:
      The value produced by the underlying thread.
    """
    try:
      return self._output_queue.get(block=block, timeout=timeout)
    except queue.Empty:
      return None

  def write(self, value, block: bool = True, timeout: Optional[float] = None):
    """'Public' method for clients to write values _to_ this thread.

    Args:
      value: The value to send to the underlying thread.
      block: Whether the client should block.
      timeout: Timeout for setting input in the queue, in seconds.
    Returns:
      The value produced by the underlying thread.
    """
    self._input_queue.put(value, block=block, timeout=timeout)

  @abc.abstractmethod
  def main(self):
    """main() function that subclasses need to override."""
    pass

  def kill(self):
    """Shorthand for clients to terminate this thread."""
    logging.info('Killing %s thread', self._name)
    # Sending a kill signal to clean up blocked read_values.
    self.write(ThreadFunction.Signal.KILL, block=True)
    # Stopping the _run loop.
    self._should_run = False

  def _run(self):
    """'Private' method that reruns main() until explicit termination."""
    logging.info('Starting %s thread.', self._name)
    while self._should_run:
      self.main()
    logging.info('Finished %s thread.', self._name)

  def _read_value(self):
    """'Protected' method for subclasses to read values from their input."""
    try:
      return self._input_queue.get(block=self._block_input)
    except queue.Empty:
      pass  # Ignore empty queues. Keep going.

  def _write_value(self, value):
    """'Protected' method for subclasses to write values to their output."""
    self._output_queue.put(value, block=self._block_output)



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The different kinds of actions that AndroidEnv supports."""

import enum


class ActionType(enum.IntEnum):
  """Integer values to describe each supported action in AndroidEnv."""
  TOUCH = 0
  LIFT = 1
  REPEAT = 2
  TEXT = 3



# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created by Danyang Zhang @X-Lance.
"""

from android_env.environment import AndroidEnv
from android_env.wrappers import base_wrapper

from typing import Dict, List, Pattern, Tuple
from typing import Union
from numbers import Number
import dm_env
from dm_env import specs
from android_env.components import action_type

import enum
from transformers import PreTrainedTokenizer
import numpy as np
import lxml.etree
import re

#import logging

def filter_elements(tree: lxml.etree.Element)\
        -> Tuple[ List[lxml.etree.Element]
                , List[List[int]]
                ]:
    #  function filter_elements {{{ # 
    node_list: List[lxml.etree.Element] = []
    bbox_list: List[List[int]] = []

    for n in tree.iterdescendants():
        n.set( "clickable"
             , str(  n.get("clickable")=="true"\
                  or n.getparent().get("clickable")=="true"
                  ).lower()
             )
        if n.get("bounds")=="[0,0][0,0]":
            continue
        if len(list(n))==0:
            bounds_match = VhIoWrapper.bounds_pattern.match(n.get("bounds"))
            bbox: List[int] = list( map( int
                                       , bounds_match.groups()
                                       )
                                  )
            if bbox[0]>=bbox[2] or bbox[1]>=bbox[3]:
                continue

            node_list.append(n)
            bbox_list.append(bbox)

    return node_list, bbox_list
    #  }}} function filter_elements # 

class VhIoWrapper(base_wrapper.BaseWrapper):
    """
    Wraps the environment with view hierarchy (VH) element iteractions.

    The actions are formalated as

    ```haskell
    data Action = CLICK Int -- eid
                | INPUT Int -- eid
                        String -- text
                | SCROLL Direction
                | GOBACK
                | NOTHING
    data Direction = UP | DOWN | LEFT | RIGHT
    ```
    """

    #  Enum Types for VH Actions {{{ # 
    class ActionType(enum.IntEnum):
        NOTHING = 0
        CLICK = 1
        INPUT = 2
        SCROLL = 3
        GOBACK = 4

    class ScrollDirection(enum.IntEnum):
        LEFT = 0
        UP = 1
        RIGHT = 2
        DOWN = 3
    #  }}} Enum Types for VH Actions # 

    bounds_pattern: Pattern[str] = re.compile(r"\[(\d+),(\d+)\]\[(\d+),(\d+)\]")

    def __init__( self
                , env: AndroidEnv
                , tokenizer: PreTrainedTokenizer
                , nb_click_frames: int = 3
                , nb_scroll_frmaes: int = 10
                ):
        #  method __init__ {{{ # 
        """
        Args:
            env (AndroidEnv): the environment to be wrapped
            tokenizer (PreTrainedTokenizer): tokenizer to tokenize the input
              text into the tokens from the vocabulary of the environment
            nb_click_frames (int): the duration the TOUCH action will last for
              the CLICK iteraction
            nb_scroll_frmaes (int): the duration the TOUCH action will last for
              the SCROLL iteraction
        """

        super(VhIoWrapper, self).__init__(env)

        self._bbox_list: List[List[int]] = [] # bboxes are saved as [x0, y0, x1, y1]
        self._instructions: List[str] = []

        self._tokenizer: PreTrainedTokenizer = tokenizer
        self._nb_click_frames: int = nb_click_frames
        self._nb_scroll_frames: int = nb_scroll_frmaes

        self._env_steps: int = 0

        self._action_spec: Dict[str, specs.Array] =\
                { "action_type": specs.DiscreteArray( num_values=len(VhIoWrapper.ActionType)
                                                    , name="action_type"
                                                    )
                , "element_id": specs.DiscreteArray( num_values=max(1, len(self._bbox_list))
                                                   , name="element_id"
                                                   )
                , "text": specs.StringArray( shape=()
                                           , name="text"
                                           )
                , "direction": specs.DiscreteArray( num_values=len(VhIoWrapper.ScrollDirection)
                                                  , name="direction"
                                                  )
                , "response": specs.StringArray( shape=()
                                               , name="response"
                                               )
                }

        scroll_trajectory = np.linspace(0.2, 0.8, num=self._nb_scroll_frames, dtype=np.float32)
        scroll_trajectory = np.stack( [ scroll_trajectory
                                      , np.zeros( (self._nb_scroll_frames,)
                                                , dtype=np.float32
                                                )
                                      ]
                                    , axis=1
                                    ) # (SF, 2); SF = self._nb_scroll_frames
        self._direction_trajectories: Dict[ VhIoWrapper.ScrollDirection
                                          , np.ndarray
                                          ] = { VhIoWrapper.ScrollDirection.LEFT:
                                                    scroll_trajectory
                                              , VhIoWrapper.ScrollDirection.UP:
                                                    scroll_trajectory[:, [1, 0]]
                                              , VhIoWrapper.ScrollDirection.RIGHT:
                                                    scroll_trajectory[::-1]
                                              , VhIoWrapper.ScrollDirection.DOWN:
                                                    scroll_trajectory[::-1, [1, 0]]
                                              }
        #  }}} method __init__ # 

    def action_spec(self) -> Dict[str, specs.Array]:
        #  method action_spec {{{ # 
        self._action_spec["element_id"] = specs.DiscreteArray( num_values=max(1, len(self._bbox_list))
                                                             , name="element_id"
                                                             )
        return self._action_spec
        #  }}} method action_spec # 

    def _process_action(self, action: Dict[str, np.ndarray])\
            -> List[Dict[str, np.ndarray]]:
        #  method _process_action {{{ # 
        """
        Args:
            action (Dict[str, np.ndarray]): dict like
              {
                "action_type": NOTHING | GOBACK
              } or
              {
                "action_type": CLICK
                "element_id": int
              } or
              {
                "action_type": INPUT
                "element_id": int
                "text": str
              } or
              {
                "action_type": SCROLL
                "direction": Direction
              }
              all the values in `action` are wrapped in np.ndarray.

        Return:
            List[Dict[str, np.ndarray]]: list of dict like
              {
                "action_type": np.ndarray of int32 scalar from TOUCH | LIFT | REPEAT
                "touch_position": np.ndarray of float32 with shape (2,)
              } or
              {
                "action_type": np.ndarray of int32 scalar from TEXT
                "input_token": np.ndarray of int32 scalar
              }
        """

        if action["action_type"]==VhIoWrapper.ActionType.NOTHING:
            return [{"action_type": np.array(action_type.ActionType.REPEAT, dtype=np.int32)}]
        if action["action_type"]==VhIoWrapper.ActionType.GOBACK:
            return []

        actions: List[Dict[str, np.ndarray]] = []

        if "element_id" in action:
            bbox: List[int] = self._bbox_list[action["element_id"]]
            x = (bbox[0]+bbox[2])/2./self._env.observation_spec()["pixels"].shape[1]
            y = (bbox[1]+bbox[3])/2./self._env.observation_spec()["pixels"].shape[0]

            touch_action = { "action_type": np.array( action_type.ActionType.TOUCH
                                                    , dtype=np.int32
                                                    )
                           , "touch_position": np.array( [x, y]
                                                       , dtype=np.float32
                                                       )
                           }
            lift_action = { "action_type": np.array( action_type.ActionType.LIFT
                                                   , dtype=np.int32
                                                   )
                          , "touch_position": np.array( [x, y]
                                                      , dtype=np.float32
                                                      )
                          }
            actions += [touch_action] * self._nb_click_frames
            actions.append(lift_action)

        if action["action_type"]==VhIoWrapper.ActionType.CLICK:
            return actions
        if action["action_type"]==VhIoWrapper.ActionType.INPUT:
            token_list: List[str] = self._tokenizer( action["text"].item()
                                                   , add_special_tokens=False
                                                   )["input_ids"]
            for tkn in token_list:
                actions.append( { "action_type": np.array( action_type.ActionType.TEXT
                                                         , dtype=np.int32
                                                         )
                                , "input_token": np.array(tkn, dtype=np.int32)
                                }
                              )
            return actions
        if action["action_type"]==VhIoWrapper.ActionType.SCROLL:
            for pst in self._direction_trajectories[action["direction"].item()]:
                actions.append( { "action_type": np.array( action_type.ActionType.TOUCH
                                                         , dtype=np.int32
                                                         )
                                , "touch_position": pst
                                }
                              )
            return actions
        #  }}} method _process_action # 

    def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
        #  method _process_timestep {{{ # 
        self._bbox_list: List[List[int]] = filter_elements(timestep.observation["view_hierarchy"])[1]
        return timestep
        #  }}} method _process_timestep # 

    def step(self, action: Dict[str, np.ndarray]) -> dm_env.TimeStep:
        #  method step {{{ # 
        """
        Args:
            action (Dict[str, np.ndarray): vh-element-wise action

        Returns:
            dm_env.TimeStep
        """

        actions: List[Dict[str, np.ndarray]] = self._process_action(action)
        self._env_steps += len(actions)

        total_reward = 0.
        instructions: List[str] = []
        for act in actions:
            #step_type: dm_env.StepType
            #reward: float
            #discount: float
            #observation: Dict[str, Any]
            #step_type, reward, discount, observation = self._env.step(act)
            timestep: dm_env.TimeStep = self._env.step(act)
            instructions += self._env.task_instructions()

            if timestep.reward>0.:
                total_reward += timestep.reward

            if timestep.last():
                self._instructions = instructions
                return dm_env.TimeStep( step_type=timestep.step_type
                                      , reward=total_reward
                                      , discount=timestep.discount
                                      , observation=timestep.observation
                                      )

        if action["action_type"]==VhIoWrapper.ActionType.INPUT:
            self._env._coordinator._task_manager._adb_controller.input_key("KEYCODE_ENTER")
        elif action["action_type"]==VhIoWrapper.ActionType.GOBACK:
            self._env._coordinator._task_manager._adb_controller.input_key("KEYCODE_BACK")

        appended_lift: Dict[str, np.ndarray] = { "action_type": np.array( action_type.ActionType.LIFT
                                                                        , dtype=np.int32
                                                                        )
                                               , "touch_position": np.array( [0., 0.]
                                                                           , dtype=np.float32
                                                                           )
                                               }
        if "response" in action\
                and action["response"] is not None\
                and action["response"] != "":
            appended_lift["response"] = action["response"]
        timestep = self._env.step(appended_lift)
        instructions += self._env.task_instructions()
        if timestep.reward>0.:
            total_reward += timestep.reward

        timestep: dm_env.TimeStep = self._process_timestep(timestep)

        self._instructions = instructions
        return dm_env.TimeStep( step_type=timestep.step_type
                              , reward=total_reward
                              , discount=timestep.discount
                              , observation=timestep.observation
                              )
        #  }}} method step # 

    def task_instructions(self, latest_only: bool = False) -> Union[str, List[str]]:
        if latest_only:
            return self._instructions[-1] if len(self._instructions)>0 else ""
        else:
            return self._instructions.copy()

    def _reset_state(self):
        self._instructions = []

    def android_logs(self) -> Dict[str, Number]:
        """
        Returns a dictionary of metrics logged by the environment.
        """
        logs: Dict[str, Number] = self._env.android_logs()
        logs.update({'env_steps': self._env_steps})
        return logs



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for AndroidEnv."""

from typing import Sequence, Tuple


from android_env.proto import adb_pb2
import numpy as np


def touch_position_to_pixel_position(
    touch_position: np.ndarray,
    width_height: Sequence[int],
) -> Tuple[int, int]:
  """Maps touch position in [0,1] to the corresponding pixel on the screen."""
  touch_pixels = (touch_position * width_height).astype(np.int32)
  cap_idx = lambda v, idx_len: min(v, idx_len - 1)
  return tuple(map(cap_idx, touch_pixels, width_height))


def transpose_pixels(frame: np.ndarray) -> np.ndarray:
  """Converts image from shape (H, W, C) to (W, H, C) and vice-versa."""
  return np.transpose(frame, axes=(1, 0, 2))


def orient_pixels(
    frame: np.ndarray,
    orientation: adb_pb2.AdbCall.Rotate.Orientation) -> np.ndarray:
  """Rotates screen pixels according to the given orientation."""
  if orientation == adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_90:
    frame = np.rot90(frame, k=3, axes=(0, 1))
  elif orientation == adb_pb2.AdbCall.Rotate.Orientation.PORTRAIT_180:
    frame = np.rot90(frame, k=2, axes=(0, 1))
  elif orientation == adb_pb2.AdbCall.Rotate.Orientation.LANDSCAPE_270:
    frame = np.rot90(frame, k=1, axes=(0, 1))
  return frame



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A component that parses and processes SetupSteps."""

import copy
import random
import re
import time
from typing import Any, Dict, Optional, Sequence

from absl import logging
from android_env.components import adb_controller as adb_control
from android_env.components import app_screen_checker
from android_env.components import errors
from android_env.components import logcat_thread
from android_env.components import event_listeners
from android_env.proto import adb_pb2
from android_env.proto import task_pb2


class SetupStepInterpreter():
  """An interpreter for SetupSteps."""

  def __init__(self,
               adb_controller: adb_control.AdbController,
               logcat: logcat_thread.LogcatThread):
    """Initializes this interpreter.

    Args:
      adb_controller: An object to communicate with Android via ADB.
      logcat: A LogcatThread instance connected to the same Android simulator
        as AdbController.
    """
    self._adb_controller = adb_controller
    self._logcat_thread = logcat
    self._last_activity = ''
    self._log_dict = {
        'error_count_adb_call': 0,
        'error_count_wait_for_app_screen': 0,
        'error_count_check_install': 0,
        'error_count_wait_for_message': 0,
        'total_time_waiting_for_app_screen': 0
    }

  def log_dict(self) -> Dict[str, Any]:
    return copy.deepcopy(self._log_dict)

  def interpret(self, setup_steps: Sequence[task_pb2.SetupStep]) -> None:
    """Returns True if parsing and processing `setup_steps` is successful."""
    if setup_steps:
      #logging.info('Executing setup steps: %s', setup_steps)
      for step in setup_steps:
        self._process_step_command(step)
      #logging.info('Done executing setup steps.')

  def _process_step_command(self, step_cmd: task_pb2.SetupStep) -> None:
    """Processes a single step command from a reset or extra setup."""

    if not step_cmd:
      logging.info('Empty step_cmd')
      return

    #logging.info('Executing step_cmd: %r', step_cmd)
    step_type = step_cmd.WhichOneof('step')
    success_condition = step_cmd.success_condition
    success_check = success_condition.WhichOneof('check')
    assert step_type or success_check, (
        'At least one of step and success_condition must be defined.')

    num_tries = 0
    max_retries = max(success_condition.num_retries, 3)
    while num_tries < max_retries:

      num_tries += 1

      try:
        self._execute_step_cmd(step_cmd, step_type)
        time.sleep(0.5)
        self._check_success(success_check, success_condition)
        return

      except NotImplementedError:
        logging.exception('Not implemented error! Skipping this step command.')
        return

      except errors.AdbControllerError:
        self._log_dict['error_count_adb_call'] += 1
        logging.warning('ADB call [%r] has failed. Try %d of %d.',
                        step_cmd.adb_call, num_tries, max_retries)

      except errors.WaitForAppScreenError:
        self._log_dict['error_count_wait_for_app_screen'] += 1
        logging.warning('Failed to wait for app screen. Try %d of %d.',
                        num_tries, max_retries)

      except errors.WaitForMessageError:
        self._log_dict['error_count_wait_for_message'] += 1
        logging.warning('Failed to wait for message. Try %d of %d.',
                        num_tries, max_retries)

      except errors.CheckInstallError:
        self._log_dict['error_count_check_install'] += 1
        logging.warning('Package [%r] not installed. Try %d of %d.',
                        success_condition.check_install.package_name,
                        num_tries, max_retries)

    raise errors.StepCommandError('Step failed: [%r]' % step_cmd)

  def _execute_step_cmd(self,
                        step_cmd: task_pb2.SetupStep,
                        step_type: Optional[str]) -> None:
    """Executes a step command of given type."""

    if not step_type:
      return

    if step_type == 'sleep':
      time.sleep(step_cmd.sleep.time_sec)
    elif step_type == 'adb_call':
      self._parse_adb_call(step_cmd.adb_call)
    else:
      raise NotImplementedError('No step command of type [%s].' % step_type)

  def _check_success(self,
                     success_check: Optional[str],
                     success_condition: task_pb2.SuccessCondition) -> None:
    """Checks whether the given success condition was met."""

    if not success_check:
      return

    if success_check == 'wait_for_app_screen':
      self._wait_for_app_screen(success_condition.wait_for_app_screen)
    elif success_check == 'check_install':
      self._check_install(success_condition.check_install)
    elif success_check == 'wait_for_message':
      self._wait_for_message(success_condition.wait_for_message)
    else:
      raise NotImplementedError('No success check called [%s].' % success_check)

  def _parse_adb_call(self, adb_cmd: adb_pb2.AdbCall) -> None:
    """Parses an adb command into set of allowed calls."""

    call_type = adb_cmd.WhichOneof('command')
    logging.info('Parsing ADB call of type: %s', call_type)

    if call_type == 'tap':
      tap = adb_cmd.tap
      self._adb_controller.input_tap(tap.x, tap.y)

    elif call_type == 'rotate':
      self._adb_controller.rotate_device(adb_cmd.rotate.orientation)

    elif call_type == 'press_button':
      if (adb_cmd.press_button.button ==
          adb_pb2.AdbCall.PressButton.Button.HOME):
        self._adb_controller.input_key('KEYCODE_HOME')
      elif (adb_cmd.press_button.button ==
            adb_pb2.AdbCall.PressButton.Button.BACK):
        self._adb_controller.input_key('KEYCODE_BACK')

    elif call_type == 'start_random_activity':
      self._last_activity = random.choice(
          list(adb_cmd.start_random_activity.activity_list))
      logging.info('Random activity: %s', self._last_activity)
      self._adb_controller.start_activity(
          self._last_activity, list(adb_cmd.start_random_activity.extra_args))
      wait_proto = task_pb2.WaitForAppScreen()
      wait_proto.app_screen.activity = self._last_activity
      wait_proto.timeout_sec = adb_cmd.start_random_activity.timeout_sec
      self._wait_for_app_screen(wait_proto)

    elif call_type == 'start_activity':
      self._adb_controller.start_activity(
          adb_cmd.start_activity.full_activity,
          list(adb_cmd.start_activity.extra_args))

    elif call_type == 'start_intent':
      self._adb_controller.start_intent(
          action=adb_cmd.start_intent.action,
          data_uri=adb_cmd.start_intent.data_uri,
          package_name=adb_cmd.start_intent.package_name)

    elif call_type == 'force_stop':
      self._adb_controller.force_stop(adb_cmd.force_stop.package_name)

    elif call_type == 'force_stop_random_activity':
      if self._last_activity:
        package_name = self._last_activity.split('/')[0]
        logging.info('Force stop package (%s)', package_name)
        self._adb_controller.force_stop(package_name)

    elif call_type == 'clear_cache':
      self._adb_controller.clear_cache(adb_cmd.clear_cache.package_name)

    elif call_type == 'clear_cache_random_activity':
      if self._last_activity:
        package_name = self._last_activity.split('/')[0]
        logging.info('Clear cache package (%s)', package_name)
        self._adb_controller.force_stop(package_name)

    elif call_type == 'grant_permissions':
      self._adb_controller.grant_permissions(
          adb_cmd.grant_permissions.package_name,
          adb_cmd.grant_permissions.permissions)

    elif call_type == 'install_apk':
      install_apk_cmd = adb_cmd.install_apk
      location_type = install_apk_cmd.WhichOneof('location')
      logging.info('location_type: %s', location_type)
      if location_type == 'filesystem':
        self._adb_controller.install_apk(install_apk_cmd.filesystem.path)
      else:
        logging.error('Unsupported location type: %r', install_apk_cmd)

    elif call_type == 'uninstall_package':
      uninstall_package_cmd = adb_cmd.uninstall_package
      self._adb_controller.uninstall_package(uninstall_package_cmd.package_name)

    elif call_type == 'start_accessibility_service':
      self._adb_controller.start_accessibility_service(
          adb_cmd.start_accessibility_service.full_service)

    elif call_type == 'start_screen_pinning':
      self._adb_controller.start_screen_pinning(
          adb_cmd.start_screen_pinning.full_activity)

    elif call_type == 'disable_animations':
      self._adb_controller.disable_animations()

    else:
      raise NotImplementedError('No ADB call type [%s].' % call_type)

  def _wait_for_app_screen(
      self, wait_for_app_screen: task_pb2.WaitForAppScreen) -> None:
    """Waits for a given `app_screen` to be the current screen."""

    logging.info('Waiting for app screen...')
    app_screen = wait_for_app_screen.app_screen
    screen_checker = app_screen_checker.AppScreenChecker(
        self._adb_controller, app_screen)

    start_time = time.time()
    while time.time() - start_time < wait_for_app_screen.timeout_sec:
      if (screen_checker.matches_current_app_screen() ==
          app_screen_checker.AppScreenChecker.Outcome.SUCCESS):
        wait_time = time.time() - start_time
        self._log_dict['total_time_waiting_for_app_screen'] += wait_time
        logging.info('Successfully waited for app screen in %r seconds: [%r]',
                     wait_time, app_screen)
        return
      time.sleep(0.1)

    wait_time = time.time() - start_time
    self._log_dict['total_time_waiting_for_app_screen'] += wait_time
    logging.error('Failed to wait for app screen in %r seconds: [%r].',
                  wait_time, app_screen)

    raise errors.WaitForAppScreenError()

  def _wait_for_message(
      self, wait_for_message: task_pb2.WaitForMessage) -> None:
    """Waits for a given message in logcat."""

    message = wait_for_message.message
    logging.info('Waiting for message: %s...', message)
    #event = re.compile(message) # zdy
    #got_message = False # zdy

    # zdy
    #def f(ev, match):
      #del ev, match
      #nonlocal got_message
      #got_message = True

    #listener = logcat_thread.EventListener(regexp=event, handler_fn=f)
    listener = event_listeners.LogEvent([], message) # zdy
    self._logcat_thread.add_event_listener(listener)
    self._logcat_thread.wait(
        event=listener, timeout_sec=wait_for_message.timeout_sec) # zdy
        #event=event, timeout_sec=wait_for_message.timeout_sec)
    self._logcat_thread.remove_event_listener(listener)
    #if got_message:
    if listener.is_set(): # zdy
      logging.info('Message received: [%r]', message)
    else:
      logging.error('Failed to wait for message: [%r].', message)
      raise errors.WaitForMessageError()

  def _check_install(self, check_install: task_pb2.CheckInstall) -> None:
    """Checks that the given package is installed."""

    package = check_install.package_name
    logging.info('Checking if package is installed: [%r]', package)

    start_time = time.time()
    while time.time() - start_time < check_install.timeout_sec:
      if self._adb_controller.is_package_installed(package):
        logging.info('Done confirming that package is installed.')
        return
      time.sleep(0.1)

    logging.error('Package not found.')
    raise errors.CheckInstallError()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.logcat_thread."""

import re
import threading
from typing import Match, Pattern

from absl.testing import absltest
from android_env.components import log_stream
from android_env.components import logcat_thread
from android_env.proto import task_pb2


class FakeStream():
  """This class simulates the logs coming from ADB."""

  def __init__(self):
    self._values = []
    self._kill = False
    self._lock = threading.Lock()

  def send_value(self, value):
    with self._lock:
      self._values.append(value)

  def has_next_value(self):
    return bool(self._values)

  def kill(self):
    self._kill = True

  def __iter__(self):
    while True:
      if self._kill:
        return
      if not self._values:
        continue
      else:
        with self._lock:
          next_value = self._values.pop(0)
        yield next_value


def _log_parsing_config():
  """Returns log_parsing_config object for testing."""

  return task_pb2.LogParsingConfig(
      filters=['AndroidRLTask:V'])


def make_stdout(data):
  """Returns a valid log output with given data as message."""
  return '         1553110400.424  5583  5658 D Tag: %s' % data


class FakeLogStream(log_stream.LogStream):
  """Add doc string."""

  def __init__(self):
    self.logs = FakeStream()
    self.stream_is_alive = True
    self._verbose = False

  def _get_stream_output(self):
    return self.logs

  def stop_stream(self):
    self.stream_is_alive = False
    self.logs.kill()


class LogcatThreadTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.fake_log_stream = FakeLogStream()

  def tearDown(self):
    self.fake_log_stream.stop_stream()
    super().tearDown()

  def test_set_filters(self):
    _ = logcat_thread.LogcatThread(
        log_stream=self.fake_log_stream,
        log_parsing_config=_log_parsing_config())
    expected_filters = ['AndroidRLTask:V', '*:S']
    self.assertEqual(expected_filters, self.fake_log_stream._filters)

  def test_kill(self):
    logcat = logcat_thread.LogcatThread(
        log_stream=self.fake_log_stream,
        log_parsing_config=_log_parsing_config())
    self.assertTrue(self.fake_log_stream.stream_is_alive)
    logcat.kill()
    self.assertFalse(self.fake_log_stream.stream_is_alive)

  def test_listeners(self):
    """Ensures that we can wait for a specific message without polling."""
    logcat = logcat_thread.LogcatThread(
        log_stream=self.fake_log_stream,
        log_parsing_config=_log_parsing_config())

    # Set up a listener that modifies an arbitrary state.
    some_state = False

    def my_handler(event: Pattern[str], match: Match[str]):
      del event, match
      nonlocal some_state
      some_state = True

    # Create a desired event and hook up the listener.
    my_event = re.compile('Hello world')
    listener = logcat_thread.EventListener(my_event, my_handler)
    logcat.add_event_listener(listener)
    self.fake_log_stream.logs.send_value('Hi there!')  # This should not match.
    self.assertFalse(some_state)
    self.fake_log_stream.logs.send_value(make_stdout('Hello world'))
    logcat.wait(event=my_event, timeout_sec=1.0)
    self.assertTrue(some_state)

    # Waiting for any events should also trigger the listener.
    some_state = False
    self.fake_log_stream.logs.send_value(make_stdout('Hello world'))
    logcat.wait(event=None, timeout_sec=1.0)
    self.assertTrue(some_state)

    # After removing the listener, it should not be called anymore.
    some_state = False
    logcat.remove_event_listener(listener)
    self.fake_log_stream.logs.send_value(make_stdout('Hello world'))
    logcat.wait(event=my_event, timeout_sec=1.0)
    self.assertFalse(some_state)


if __name__ == '__main__':
  absltest.main()
